"""
This module defines the database models for all in-game objects, that
is, all objects that has an actual existence in-game.

Each database object is 'decorated' with a 'typeclass', a normal
python class that implements all the various logics needed by the game
in question. Objects created of this class transparently communicate
with its related database object for storing all attributes. The
admin should usually not have to deal directly with this database
object layer.

Attributes are separate objects that store values persistently onto
the database object. Like everything else, they can be accessed
transparently through the decorating TypeClass.
"""
from collections import defaultdict

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import validate_comma_separated_integer_list
from django.db import models

from evennia.objects.manager import ObjectDBManager
from evennia.typeclasses.models import TypedObject
from evennia.utils import logger
from evennia.utils.utils import dbref, lazy_property, make_iter


class ContentsHandler:
    """
    Handles and caches the contents of an object to avoid excessive
    lookups (this is done very often due to cmdhandler needing to look
    for object-cmdsets). It is stored on the 'contents_cache' property
    of the ObjectDB.
    """

    def __init__(self, obj):
        """
        Sets up the contents handler.

        Args:
            obj (Object):  The object on which the
                handler is defined

        Notes:
            This was changed from using `set` to using `dict` internally
            in order to retain insertion order.

        """
        self.obj = obj
        self._pkcache = {}
        self._idcache = obj.__class__.__instance_cache__
        self._typecache = defaultdict(dict)
        self.init()

    def load(self):
        """
        Retrieves all objects from database. Used for initializing.

        Returns:
            Objects (list of ObjectDB)
        """
        return list(self.obj.locations_set.all())

    def init(self):
        """
        Re-initialize the content cache

        """
        objects = self.load()
        self._pkcache = {obj.pk: True for obj in objects}
        for obj in objects:
            for ctype in obj._content_types:
                self._typecache[ctype][obj.pk] = True

    def get(self, exclude=None, content_type=None):
        """
        Return the contents of the cache.

        Args:
            exclude (Object or list of Object): object(s) to ignore
            content_type (str or None): Filter list by a content-type. If None, don't filter.

        Returns:
            objects (list): the Objects inside this location

        """
        if content_type is not None:
            pks = self._typecache[content_type].keys()
        else:
            pks = self._pkcache.keys()
        if exclude:
            pks = set(pks) - {excl.pk for excl in make_iter(exclude)}
        try:
            return [self._idcache[pk] for pk in pks]
        except KeyError:
            # this can happen if the idmapper cache was cleared for an object
            # in the contents cache. If so we need to re-initialize and try again.
            self.init()
            try:
                return [self._idcache[pk] for pk in pks]
            except KeyError:
                # this means an actual failure of caching. Return real database match.
                logger.log_err("contents cache failed for %s." % self.obj.key)
                return self.load()

    def add(self, obj):
        """
        Add a new object to this location

        Args:
            obj (Object): object to add

        """
        self._pkcache[obj.pk] = obj
        for ctype in obj._content_types:
            self._typecache[ctype][obj.pk] = True

    def remove(self, obj):
        """
        Remove object from this location

        Args:
            obj (Object): object to remove

        """
        self._pkcache.pop(obj.pk, None)
        for ctype in obj._content_types:
            if obj.pk in self._typecache[ctype]:
                self._typecache[ctype].pop(obj.pk, None)

    def clear(self):
        """
        Clear the contents cache and re-initialize

        """
        self._pkcache = {}
        self._typecache = defaultdict(dict)
        self.init()


# -------------------------------------------------------------
#
# ObjectDB
#
# -------------------------------------------------------------


class ObjectDB(TypedObject):
    """
    All objects in the game use the ObjectDB model to store
    data in the database. This is handled transparently through
    the typeclass system.

    Note that the base objectdb is very simple, with
    few defined fields. Use attributes to extend your
    type class with new database-stored variables.

    The TypedObject supplies the following (inherited) properties:

      - key - main name
      - name - alias for key
      - db_typeclass_path - the path to the decorating typeclass
      - db_date_created - time stamp of object creation
      - permissions - perm strings
      - locks - lock definitions (handler)
      - dbref - #id of object
      - db - persistent attribute storage
      - ndb - non-persistent attribute storage

    The ObjectDB adds the following properties:

      - account - optional connected account (always together with sessid)
      - sessid - optional connection session id (always together with account)
      - location - in-game location of object
      - home - safety location for object (handler)
      - scripts - scripts assigned to object (handler from typeclass)
      - cmdset - active cmdset on object (handler from typeclass)
      - aliases - aliases for this object (property)
      - nicks - nicknames for *other* things in Evennia (handler)
      - sessions - sessions connected to this object (see also account)
      - has_account - bool if an active account is currently connected
      - contents - other objects having this object as location
      - exits - exits from this object

    """

    #
    # ObjectDB Database model setup
    #
    #
    # inherited fields (from TypedObject):
    # db_key (also 'name' works), db_typeclass_path, db_date_created,
    # db_permissions
    #
    # These databse fields (including the inherited ones) should normally be
    # managed by their corresponding wrapper properties, named same as the
    # field, but without the db_* prefix (e.g. the db_key field is set with
    # self.key instead). The wrappers are created at the metaclass level and
    # will automatically save and cache the data more efficiently.

    # If this is a character object, the account is connected here.
    db_account = models.ForeignKey(
        "accounts.AccountDB",
        null=True,
        verbose_name="account",
        on_delete=models.SET_NULL,
        help_text="an Account connected to this object, if any.",
    )

    # the session id associated with this account, if any
    db_sessid = models.CharField(
        null=True,
        max_length=32,
        validators=[validate_comma_separated_integer_list],
        verbose_name="session id",
        help_text="csv list of session ids of connected Account, if any.",
    )
    # The location in the game world. Since this one is likely
    # to change often, we set this with the 'location' property
    # to transparently handle Typeclassing.
    db_location = models.ForeignKey(
        "self",
        related_name="locations_set",
        db_index=True,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name="game location",
    )
    # a safety location, this usually don't change much.
    db_home = models.ForeignKey(
        "self",
        related_name="homes_set",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name="home location",
    )
    # destination of this object - primarily used by exits.
    db_destination = models.ForeignKey(
        "self",
        related_name="destinations_set",
        db_index=True,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name="destination",
        help_text="a destination, used only by exit objects.",
    )
    # database storage of persistant cmdsets.
    db_cmdset_storage = models.CharField(
        "cmdset",
        max_length=255,
        null=True,
        blank=True,
        help_text="optional python path to a cmdset class.",
    )

    # Database manager
    objects = ObjectDBManager()

    # defaults
    __settingsclasspath__ = settings.BASE_OBJECT_TYPECLASS
    __defaultclasspath__ = "evennia.objects.objects.DefaultObject"
    __applabel__ = "objects"

    @lazy_property
    def contents_cache(self):
        return ContentsHandler(self)

    # cmdset_storage property handling
    def __cmdset_storage_get(self):
        """getter"""
        storage = self.db_cmdset_storage
        return [path.strip() for path in storage.split(",")] if storage else []

    def __cmdset_storage_set(self, value):
        """setter"""
        self.db_cmdset_storage = ",".join(str(val).strip() for val in make_iter(value))
        self.save(update_fields=["db_cmdset_storage"])

    def __cmdset_storage_del(self):
        """deleter"""
        self.db_cmdset_storage = None
        self.save(update_fields=["db_cmdset_storage"])

    cmdset_storage = property(__cmdset_storage_get, __cmdset_storage_set, __cmdset_storage_del)

    # location getsetter
    def __location_get(self):
        """Get location"""
        return self.db_location

    def __location_set(self, location):
        """Set location, checking for loops and allowing dbref"""
        if isinstance(location, (str, int)):
            # allow setting of #dbref
            dbid = dbref(location, reqhash=False)
            if dbid:
                try:
                    location = ObjectDB.objects.get(id=dbid)
                except ObjectDoesNotExist:
                    # maybe it is just a name that happens to look like a dbid
                    pass
        try:

            def is_loc_loop(loc, depth=0):
                """Recursively traverse target location, trying to catch a loop."""
                if depth > 10:
                    return None
                elif loc == self:
                    raise RuntimeError
                elif loc is None:
                    raise RuntimeWarning
                return is_loc_loop(loc.db_location, depth + 1)

            try:
                is_loc_loop(location)
            except RuntimeWarning:
                # we caught an infinite location loop!
                # (location1 is in location2 which is in location1 ...)
                pass

            # if we get to this point we are ready to change location

            old_location = self.db_location

            # this is checked in _db_db_location_post_save below
            self._safe_contents_update = True

            # actually set the field (this will error if location is invalid)
            self.db_location = location
            self.save(update_fields=["db_location"])

            # remove the safe flag
            del self._safe_contents_update

            # update the contents cache
            if old_location:
                old_location.contents_cache.remove(self)
            if self.db_location:
                self.db_location.contents_cache.add(self)

        except RuntimeError:
            errmsg = "Error: %s.location = %s creates a location loop." % (self.key, location)
            raise RuntimeError(errmsg)
        except Exception:
            # raising here gives more info for now
            raise
            # errmsg = "Error (%s): %s is not a valid location." % (str(e), location)
            # raise RuntimeError(errmsg)
        return

    def __location_del(self):
        """Cleanly delete the location reference"""
        self.db_location = None
        self.save(update_fields=["db_location"])

    location = property(__location_get, __location_set, __location_del)

    def at_db_location_postsave(self, new):
        """
        This is called automatically after the location field was
        saved, no matter how. It checks for a variable
        _safe_contents_update to know if the save was triggered via
        the location handler (which updates the contents cache) or
        not.

        Args:
            new (bool): Set if this location has not yet been saved before.

        """
        if not hasattr(self, "_safe_contents_update"):
            # changed/set outside of the location handler
            if new:
                # if new, there is no previous location to worry about
                if self.db_location:
                    self.db_location.contents_cache.add(self)
            else:
                # Since we cannot know at this point was old_location was, we
                # trigger a full-on contents_cache update here.
                logger.log_warn(
                    "db_location direct save triggered contents_cache.init() for all objects!"
                )
                [o.contents_cache.init() for o in self.__dbclass__.get_all_cached_instances()]

    class Meta:
        """Define Django meta options"""

        verbose_name = "Object"
        verbose_name_plural = "Objects"



"""
Components - ChrisLR 2022

This file contains classes functions related to signals.
"""


def as_listener(func=None, signal_name=None):
    """
    Decorator style function that marks a method to be connected as listener.
    It will use the provided signal name and default to the decorated function name.

    Args:
        func (callable): The method to mark as listener
        signal_name (str): The name of the signal to listen to, defaults to function name.
    """
    if not func and signal_name:

        def wrapper(func):
            func._listener_signal_name = signal_name
            return func

        return wrapper

    signal_name = func.__name__
    func._listener_signal_name = signal_name
    return func


def as_responder(func=None, signal_name=None):
    """
    Decorator style function that marks a method to be connected as responder.
    It will use the provided signal name and default to the decorated function name.

    Args:
        func (callable): The method to mark as responder
        signal_name (str): The name of the signal to respond to, defaults to function name.
    """
    if not func and signal_name:

        def wrapper(func):
            func._responder_signal_name = signal_name
            return func

        return wrapper

    signal_name = func.__name__
    func._responder_signal_name = signal_name
    return func


class SignalsHandler(object):
    """
    This object handles all about signals.
    It holds the connected listeners and responders.
    It allows triggering signals or querying responders.
    """

    def __init__(self, host):
        self.host = host
        self.listeners = {}
        self.responders = {}
        self.add_object_listeners_and_responders(host)

    def add_listener(self, signal_name, callback):
        """
        Connect a listener to a specific signal.

        Args:
            signal_name (str): The name of the signal to listen to
            callback (callable): The callable that is called when the signal is triggered
        """

        signal_listeners = self.listeners.setdefault(signal_name, [])
        if callback not in signal_listeners:
            signal_listeners.append(callback)

    def add_responder(self, signal_name, callback):
        """
        Connect a responder to a specific signal.

        Args:
            signal_name (str): The name of the signal to respond to
            callback (callable): The callable that is called when the signal is queried
        """

        signal_responders = self.responders.setdefault(signal_name, [])
        if callback not in signal_responders:
            signal_responders.append(callback)

    def remove_listener(self, signal_name, callback):
        """
        Removes a listener for a specific signal.

        Args:
            signal_name (str): The name of the signal to disconnect from
            callback (callable): The callable that was used to connect
        """

        signal_listeners = self.listeners.get(signal_name)
        if not signal_listeners:
            return

        if callback in signal_listeners:
            signal_listeners.remove(callback)

    def remove_responder(self, signal_name, callback):
        """
        Removes a responder for a specific signal.

        Args:
            signal_name (str): The name of the signal to disconnect from
            callback (callable): The callable that was used to connect
        """
        signal_responders = self.responders.get(signal_name)
        if not signal_responders:
            return

        if callback in signal_responders:
            signal_responders.remove(callback)

    def trigger(self, signal_name, *args, **kwargs):
        """
        Triggers a specific signal with specified args and kwargs
        This method does not return anything

        Args:
            signal_name (str): The name of the signal to trigger
        """

        callbacks = self.listeners.get(signal_name)
        if not callbacks:
            return

        for callback in callbacks:
            callback(*args, **kwargs)

    def query(self, signal_name, *args, default=None, aggregate_func=None, **kwargs):
        """
        Queries a specific signal with specified args and kwargs
        This method will return the responses from its connected responders.
        If an aggregate_func is specified, it is called with the responses
        and its result is returned instead.

        Args:
            signal_name (str): The name of the signal to trigger
            default (any): The value to use when no responses are given
                           It will be passed to aggregate_func if it is also given.
            aggregate_func (callable): The function to process the results before returning.

        Returns:
            list: An iterable of the responses
                  OR the aggregated result when aggregate_func is specified.

        """
        callbacks = self.responders.get(signal_name)

        if not callbacks:
            default = [] if default is None else default
            if aggregate_func:
                return aggregate_func(default)
            return default

        responses = []
        for callback in callbacks:
            response = callback(*args, **kwargs)
            if response is not None:
                responses.append(response)

        if aggregate_func and responses:
            return aggregate_func(responses)

        return responses

    def add_object_listeners_and_responders(self, obj):
        """
        This connects the methods marked as listener or responder from an object.

        Args:
            obj (object): The instance of an object to connect to this handler.
        """
        obj_type = type(obj)
        for att_name in dir(obj_type):
            if att_name.startswith("__"):
                continue

            att_obj = getattr(obj_type, att_name, None)
            if att_obj is None:
                continue

            listener_signal_name = getattr(att_obj, "_listener_signal_name", None)
            if listener_signal_name:
                callback = getattr(obj, att_name)
                self.add_listener(signal_name=listener_signal_name, callback=callback)

            responder_signal_name = getattr(att_obj, "_responder_signal_name", None)
            if responder_signal_name:
                callback = getattr(obj, att_name)
                self.add_responder(signal_name=responder_signal_name, callback=callback)

    def remove_object_listeners_and_responders(self, obj):
        """
        This disconnects the methods marked as listener or responder from an object.

        Args:
            obj (object): The instance of an object to disconnect from this handler.
        """
        for att_name in dir(obj):
            if att_name.startswith("__"):
                continue

            att_obj = getattr(obj, att_name, None)
            if att_obj is None:
                continue

            listener_signal_name = getattr(att_obj, "_listener_signal_name", None)
            if listener_signal_name:
                callback = getattr(obj, att_name)
                self.remove_listener(signal_name=listener_signal_name, callback=callback)

            responder_signal_name = getattr(att_obj, "_responder_signal_name", None)
            if responder_signal_name:
                callback = getattr(obj, att_name)
                self.remove_responder(signal_name=responder_signal_name, callback=callback)



"""
This is a simple context factory for auto-creating
SSL keys and certificates.

"""
import os
import sys

try:
    import OpenSSL
    from twisted.internet import ssl as twisted_ssl
except ImportError as error:
    errstr = """
    {err}
    SSL requires the PyOpenSSL library:
        pip install pyopenssl
    """
    raise ImportError(errstr.format(err=error))

from django.conf import settings

from evennia.utils.utils import class_from_module

_GAME_DIR = settings.GAME_DIR

# messages

NO_AUTOGEN = """

{err}
Evennia could not auto-generate the SSL private key. If this error
persists, create {keyfile} yourself using third-party tools.
"""

NO_AUTOCERT = """

{err}
Evennia's SSL context factory could not automatically, create an SSL
certificate {certfile}.

A private key {keyfile} was already created. Please create {certfile}
manually using the commands valid  for your operating system, for
example (linux, using the openssl program):
    {exestring}
"""

_TELNET_PROTOCOL_CLASS = class_from_module(settings.TELNET_PROTOCOL_CLASS)


class SSLProtocol(_TELNET_PROTOCOL_CLASS):
    """
    Communication is the same as telnet, except data transfer
    is done with encryption.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.protocol_name = "ssl"


def verify_SSL_key_and_cert(keyfile, certfile):
    """
    This function looks for RSA key and certificate in the current
    directory. If files ssl.key and ssl.cert does not exist, they
    are created.

    """

    if not (os.path.exists(keyfile) and os.path.exists(certfile)):
        # key/cert does not exist. Create.
        import subprocess

        from Crypto.PublicKey import RSA
        from twisted.conch.ssh.keys import Key

        print("  Creating SSL key and certificate ... ", end=" ")

        try:
            # create the RSA key and store it.
            KEY_LENGTH = 2048
            rsa_key = Key(RSA.generate(KEY_LENGTH))
            key_string = rsa_key.toString(type="OPENSSH")
            with open(keyfile, "w+b") as fil:
                fil.write(key_string)
        except Exception as err:
            print(NO_AUTOGEN.format(err=err, keyfile=keyfile))
            sys.exit(5)

        # try to create the certificate
        CERT_EXPIRE = 365 * 20  # twenty years validity
        # default:
        # openssl req -new -x509 -key ssl.key -out ssl.cert -days 7300
        exestring = "openssl req -new -x509 -key %s -out %s -days %s" % (
            keyfile,
            certfile,
            CERT_EXPIRE,
        )
        try:
            subprocess.call(exestring)
        except OSError as err:
            raise OSError(
                NO_AUTOCERT.format(err=err, certfile=certfile, keyfile=keyfile, exestring=exestring)
            )
        print("done.")


def getSSLContext():
    """
    This is called by the portal when creating the SSL context
    server-side.

    Returns:
        ssl_context (tuple): A key and certificate that is either
            existing previously or or created on the fly.

    """
    keyfile = os.path.join(_GAME_DIR, "server", "ssl.key")
    certfile = os.path.join(_GAME_DIR, "server", "ssl.cert")

    verify_SSL_key_and_cert(keyfile, certfile)
    return twisted_ssl.DefaultOpenSSLContextFactory(keyfile, certfile)



"""
Functions to extend the event system.

These functions are to be used by developers to customize events and callbacks.

"""

from django.conf import settings

from evennia import ScriptDB, logger
from evennia.contrib.base_systems.custom_gametime import UNITS, gametime_to_realtime
from evennia.contrib.base_systems.custom_gametime import (
    real_seconds_until as custom_rsu,
)
from evennia.utils.create import create_script
from evennia.utils.gametime import real_seconds_until as standard_rsu
from evennia.utils.utils import class_from_module

# Temporary storage for events waiting for the script to be started
EVENTS = []


def get_event_handler():
    """Return the event handler or None."""
    try:
        script = ScriptDB.objects.get(db_key="event_handler")
    except ScriptDB.DoesNotExist:
        logger.log_trace("Can't get the event handler.")
        script = None

    return script


def register_events(path_or_typeclass):
    """
    Register the events in this typeclass.

    Args:
        path_or_typeclass (str or type): the Python path leading to the
                class containing events, or the class itself.

    Returns:
        The typeclass itself.

    Notes:
        This function will read events from the `_events` class variable
        defined in the typeclass given in parameters.  It will add
        the events, either to the script if it exists, or to some
        temporary storage, waiting for the script to be initialized.

    """
    if isinstance(path_or_typeclass, str):
        typeclass = class_from_module(path_or_typeclass)
    else:
        typeclass = path_or_typeclass

    typeclass_name = typeclass.__module__ + "." + typeclass.__name__
    try:
        storage = ScriptDB.objects.get(db_key="event_handler")
        assert storage.ndb.events is not None
    except (ScriptDB.DoesNotExist, AssertionError):
        storage = EVENTS

    # If the script is started, add the event directly.
    # Otherwise, add it to the temporary storage.
    for name, tup in getattr(typeclass, "_events", {}).items():
        if len(tup) == 4:
            variables, help_text, custom_call, custom_add = tup
        elif len(tup) == 3:
            variables, help_text, custom_call = tup
            custom_add = None
        elif len(tup) == 2:
            variables, help_text = tup
            custom_call = None
            custom_add = None
        else:
            variables = help_text = custom_call = custom_add = None

        if isinstance(storage, list):
            storage.append((typeclass_name, name, variables, help_text, custom_call, custom_add))
        else:
            storage.add_event(typeclass_name, name, variables, help_text, custom_call, custom_add)

    return typeclass


# Custom callbacks for specific event types


def get_next_wait(format):
    """
    Get the length of time in seconds before format.

    Args:
        format (str): a time format matching the set calendar.

    Returns:
        until (int or float): the number of seconds until the event.
        usual (int or float): the usual number of seconds between events.
        format (str): a string format representing the time.

    Notes:
        The time format could be something like "2018-01-08 12:00".  The
        number of units set in the calendar affects the way seconds are
        calculated.

    """
    calendar = getattr(settings, "EVENTS_CALENDAR", None)
    if calendar is None:
        logger.log_err(
            "A time-related event has been set whereas "
            "the gametime calendar has not been set in the settings."
        )
        return
    elif calendar == "standard":
        rsu = standard_rsu
        units = ["min", "hour", "day", "month", "year"]
    elif calendar == "custom":
        rsu = custom_rsu
        back = dict([(value, name) for name, value in UNITS.items()])
        sorted_units = sorted(back.items())
        del sorted_units[0]
        units = [n for v, n in sorted_units]

    params = {}
    for delimiter in ("-", ":"):
        format = format.replace(delimiter, " ")

    pieces = list(reversed(format.split()))
    details = []
    i = 0
    for uname in units:
        try:
            piece = pieces[i]
        except IndexError:
            break

        if not piece.isdigit():
            logger.log_trace(
                "The time specified '{}' in {} isn't " "a valid number".format(piece, format)
            )
            return

        # Convert the piece to int
        piece = int(piece)
        params[uname] = piece
        details.append("{}={}".format(uname, piece))
        if i < len(units):
            next_unit = units[i + 1]
        else:
            next_unit = None
        i += 1

    params["sec"] = 0
    details = " ".join(details)
    until = rsu(**params)
    usual = -1
    if next_unit:
        kwargs = {next_unit: 1}
        usual = gametime_to_realtime(**kwargs)
    return until, usual, details


def time_event(obj, event_name, number, parameters):
    """
    Create a time-related event.

    Args:
        obj (Object): the object on which sits the event.
        event_name (str): the event's name.
        number (int): the number of the event.
        parameters (str): the parameter of the event.

    """
    seconds, usual, key = get_next_wait(parameters)
    script = create_script(
        "evennia.contrib.base_systems.ingame_python.scripts.TimeEventScript",
        interval=seconds,
        obj=obj,
    )
    script.key = key
    script.desc = "event on {}".format(key)
    script.db.time_format = parameters
    script.db.number = number
    script.ndb.usual = usual


def keyword_event(callbacks, parameters):
    """
    Custom call for events with keywords (like push, or pull, or turn...).

    Args:
        callbacks (list of dict): the list of callbacks to be called.
        parameters (str): the actual parameters entered to trigger the callback.

    Returns:
        A list containing the callback dictionaries to be called.

    Notes:
        This function should be imported and added as a custom_call
        parameter to add the event when the event supports keywords
        as parameters.  Keywords in parameters are one or more words
        separated by a comma.  For instance, a 'push 1, one' callback can
        be set to trigger when the player 'push 1' or 'push one'.

    """
    key = parameters.strip().lower()
    to_call = []
    for callback in callbacks:
        keys = callback["parameters"]
        if not keys or key in [p.strip().lower() for p in keys.split(",")]:
            to_call.append(callback)

    return to_call


def phrase_event(callbacks, parameters):
    """
    Custom call for events with keywords in sentences (like say or whisper).

    Args:
        callbacks (list of dict): the list of callbacks to be called.
        parameters (str): the actual parameters entered to trigger the callback.

    Returns:
        A list containing the callback dictionaries to be called.

    Notes:
        This function should be imported and added as a custom_call
        parameter to add the event when the event supports keywords
        in phrases as parameters.  Keywords in parameters are one or more
        words separated by a comma.  For instance, a 'say yes, okay' callback
        can be set to trigger when the player says something containing
        either "yes" or "okay" (maybe 'say I don't like it, but okay').

    """
    phrase = parameters.strip().lower()
    # Remove punctuation marks
    punctuations = ',.";?!'
    for p in punctuations:
        phrase = phrase.replace(p, " ")
    words = phrase.split()
    words = [w.strip("' ") for w in words if w.strip("' ")]
    to_call = []
    for callback in callbacks:
        keys = callback["parameters"]
        if not keys or any(key.strip().lower() in words for key in keys.split(",")):
            to_call.append(callback)

    return to_call


class InterruptEvent(RuntimeError):

    """
    Interrupt the current event.

    You shouldn't have to use this exception directly, probably use the
    `deny()` function that handles it instead.

    """

    pass



"""
Monitors - catch changes to model fields and Attributes.

The MONITOR_HANDLER singleton from this module offers the following
functionality:

- Field-monitor - track a object's specific database field and perform
    an action whenever that field *changes* for whatever reason.
- Attribute-monitor tracks an object's specific Attribute and perform
    an action whenever that Attribute *changes* for whatever reason.

"""
import inspect
from collections import defaultdict

from evennia.server.models import ServerConfig
from evennia.utils import logger, variable_from_module
from evennia.utils.dbserialize import dbserialize, dbunserialize

_SA = object.__setattr__
_GA = object.__getattribute__
_DA = object.__delattr__


class MonitorHandler(object):
    """
    This is a resource singleton that allows for registering
    callbacks for when a field or Attribute is updated (saved).

    """

    def __init__(self):
        """
        Initialize the handler.

        """
        self.savekey = "_monitorhandler_save"
        self.monitors = defaultdict(lambda: defaultdict(dict))

    def save(self):
        """
        Store our monitors to the database. This is called
        by the server process.

        Since dbserialize can't handle defaultdicts, we convert to an
        intermediary save format ((obj,fieldname, idstring, callback, kwargs), ...)

        """
        savedata = []
        if self.monitors:
            for obj in self.monitors:
                for fieldname in self.monitors[obj]:
                    for idstring, (callback, persistent, kwargs) in self.monitors[obj][
                        fieldname
                    ].items():
                        path = "%s.%s" % (callback.__module__, callback.__name__)
                        savedata.append((obj, fieldname, idstring, path, persistent, kwargs))
            savedata = dbserialize(savedata)
            ServerConfig.objects.conf(key=self.savekey, value=savedata)

    def restore(self, server_reload=True):
        """
        Restore our monitors after a reload. This is called
        by the server process.

        Args:
            server_reload (bool, optional): If this is False, it means
                the server went through a cold reboot and all
                non-persistent tickers must be killed.

        """
        self.monitors = defaultdict(lambda: defaultdict(dict))
        restored_monitors = ServerConfig.objects.conf(key=self.savekey)
        if restored_monitors:
            restored_monitors = dbunserialize(restored_monitors)
            for obj, fieldname, idstring, path, persistent, kwargs in restored_monitors:
                try:
                    if not server_reload and not persistent:
                        # this monitor will not be restarted
                        continue
                    if "session" in kwargs and not kwargs["session"]:
                        # the session was removed because it no longer
                        # exists. Don't restart the monitor.
                        continue
                    modname, varname = path.rsplit(".", 1)
                    callback = variable_from_module(modname, varname)

                    if obj and hasattr(obj, fieldname):
                        self.monitors[obj][fieldname][idstring] = (callback, persistent, kwargs)
                except Exception:
                    continue
        # make sure to clean data from database
        ServerConfig.objects.conf(key=self.savekey, delete=True)

    def _attr_category_fieldname(self, fieldname, category):
        """
        Modify the saved fieldname to make sure to differentiate between Attributes
        with different categories.

        """
        return f"{fieldname}[{category}]" if category else fieldname

    def at_update(self, obj, fieldname):
        """
        Called by the field/attribute as it saves.

        """
        # if this an Attribute with a category we should differentiate
        fieldname = self._attr_category_fieldname(
            fieldname,
            obj.db_category if fieldname == "db_value" and hasattr(obj, "db_category") else None,
        )

        to_delete = []
        if obj in self.monitors and fieldname in self.monitors[obj]:
            for idstring, (callback, persistent, kwargs) in self.monitors[obj][fieldname].items():
                try:
                    callback(obj=obj, fieldname=fieldname, **kwargs)
                except Exception:
                    to_delete.append((obj, fieldname, idstring))
                    logger.log_trace("Monitor callback was removed.")
        # we cleanup non-found monitors (has to be done after loop)
        for obj, fieldname, idstring in to_delete:
            del self.monitors[obj][fieldname][idstring]

    def add(self, obj, fieldname, callback, idstring="", persistent=False, category=None, **kwargs):
        """
        Add monitoring to a given field or Attribute. A field must
        be specified with the full db_* name or it will be assumed
        to be an Attribute (so `db_key`, not just `key`).

        Args:
            obj (Typeclassed Entity): The entity on which to monitor a
                field or Attribute.
            fieldname (str): Name of field (db_*) or Attribute to monitor.
            callback (callable): A callable on the form `callable(**kwargs),
                where kwargs holds keys fieldname and obj.
            idstring (str, optional): An id to separate this monitor from other monitors
                of the same field and object.
            persistent (bool, optional): If False, the monitor will survive
                a server reload but not a cold restart. This is default.
            category (str, optional): This is only used if `fieldname` refers to
                an Attribute (i.e. it does not start with `db_`). You must specify this
                if you want to target an Attribute with a category.

        Keyword Args:
            session (Session): If this keyword is given, the monitorhandler will
                correctly analyze it and remove the monitor if after a reload/reboot
                the session is no longer valid.
            any (any): Any other kwargs are passed on to the callback. Remember that
                all kwargs must be possible to pickle!

        """
        if not fieldname.startswith("db_") or not hasattr(obj, fieldname):
            # an Attribute - we track its db_value field
            obj = obj.attributes.get(fieldname, category=category, return_obj=True)
            if not obj:
                return
            fieldname = self._attr_category_fieldname("db_value", category)

        # we try to serialize this data to test it's valid. Otherwise we won't accept it.
        try:
            if not inspect.isfunction(callback):
                raise TypeError("callback is not a function.")
            dbserialize((obj, fieldname, callback, idstring, persistent, kwargs))
        except Exception:
            err = "Invalid monitor definition: \n" " (%s, %s, %s, %s, %s, %s)" % (
                obj,
                fieldname,
                callback,
                idstring,
                persistent,
                kwargs,
            )
            logger.log_trace(err)
        else:
            self.monitors[obj][fieldname][idstring] = (callback, persistent, kwargs)

    def remove(self, obj, fieldname, idstring="", category=None):
        """
        Remove a monitor.
        """
        if not fieldname.startswith("db_") or not hasattr(obj, fieldname):
            obj = obj.attributes.get(fieldname, return_obj=True)
            if not obj:
                return
            fieldname = self._attr_category_fieldname("db_value", category)

        idstring_dict = self.monitors[obj][fieldname]
        if idstring in idstring_dict:
            del self.monitors[obj][fieldname][idstring]

    def clear(self):
        """
        Delete all monitors.
        """
        self.monitors = defaultdict(lambda: defaultdict(dict))

    def all(self, obj=None):
        """
        List all monitors or all monitors of a given object.

        Args:
            obj (Object): The object on which to list all monitors.

        Returns:
            monitors (list): The handled monitors.

        """
        output = []
        objs = [obj] if obj else self.monitors

        for obj in objs:
            for fieldname in self.monitors[obj]:
                for idstring, (callback, persistent, kwargs) in self.monitors[obj][
                    fieldname
                ].items():
                    output.append((obj, fieldname, idstring, persistent, kwargs))
        return output


# access object
MONITOR_HANDLER = MonitorHandler()



#
# Batchcode script
#
#
# The Batch-code processor accepts full Python modules (e.g. "batch.py") that
# look identical to normal Python files with a few exceptions that allows them
# to be executed in blocks. This way of working assures a sequential execution
# of the file and allows for features like stepping from block to block
# (without executing those coming before), as well as automatic deletion
# of created objects etc. You can however also run a batch-code python file
# directly using Python.

# Code blocks are separated by python comments starting with special code words.

# #HEADER - this denotes commands global to the entire file, such as
#           import statements and global variables. They will
#           automatically be made available for each block.  Observe
#           that changes to these variables made in one block are not
#           preserved between blocks!)
# #CODE (infotext) [objname, objname, ...] - This designates a code block that
#            will be executed like a stand-alone piece of code together with
#            any #HEADER defined.
#            infotext is a describing text about what goes on in this block.
#            It will be shown by the batch-processing command.
#            <objname>s mark the (variable-)names of objects created in
#            the code, and which may be auto-deleted by the processor if
#            desired (such as when debugging the script). E.g., if the code
#            contains the command myobj = create.create_object(...), you could
#            put 'myobj' in the #CODE header regardless of what the created
#            object is actually called in-game.
# #INSERT filename - this includes another code batch file into this one. The
#            named file will be loaded and run at the position of the #INSERT.
#            Note that the inserted file will use its own #HEADERs and not
#            have access to the #HEADERs of the inserting file.

# The following variable is automatically made available for the script:

# caller - the object executing the script
#


# HEADER

# everything in this block will be appended to the beginning of
# all other #CODE blocks when they are executed.

from evennia import DefaultObject, create_object, search_object
from evennia.contrib.tutorials import red_button

limbo = search_object("Limbo")[0]


# CODE

# This is the first code block. Within each block, Python
# code works as normal. Note how we make use if imports and
# 'limbo' defined in the #HEADER block. This block's header
# offers no information about red_button variable, so it
# won't be able to be deleted in debug mode.

# create a red button in limbo
red_button = create_object(
    red_button.RedButton, key="Red button", location=limbo, aliases=["button"]
)

# we take a look at what we created
caller.msg("A %s was created." % red_button.key)

# CODE

# this code block has 'table' and 'chair' set as deletable
# objects. This means that when the batchcode processor runs in
# testing mode, objects created in these variables will be deleted
# again (so as to avoid duplicate objects when testing the script many
# times).

# the Python variables we assign to must match the ones given in the
# header for the system to be able to delete them afterwards during a
# debugging run.
table = create_object(DefaultObject, key="Table", location=limbo)
chair = create_object(DefaultObject, key="Chair", location=limbo)

string = "A %s and %s were created."
if DEBUG:
    string += " Since debug was active, they were deleted again."
    table.delete()
    chair.delete()

caller.msg(string % (table, chair))



"""
Gendersub

Griatch 2015

This is a simple gender-aware Character class for allowing users to
insert custom markers in their text to indicate gender-aware
messaging. It relies on a modified msg() and is meant as an
inspiration and starting point to how to do stuff like this.

An object can have the following genders:
 - male (he/his)
 - female (her/hers)
 - neutral (it/its)
 - ambiguous (they/them/their/theirs)

Usage

When in use, messages can contain special tags to indicate pronouns gendered
based on the one being addressed. Capitalization will be retained.

- `|s`, `|S`: Subjective form: he, she, it, He, She, It, They
- `|o`, `|O`: Objective form: him, her, it, Him, Her, It, Them
- `|p`, `|P`: Possessive form: his, her, its, His, Her, Its, Their
- `|a`, `|A`: Absolute Possessive form: his, hers, its, His, Hers, Its, Theirs

For example,

```
char.msg("%s falls on |p face with a thud." % char.key)
"Tom falls on his face with a thud"
```

The default gender is "ambiguous" (they/them/their/theirs).

To use, have DefaultCharacter inherit from this, or change
setting.DEFAULT_CHARACTER to point to this class.

The `gender` command is used to set the gender. It needs to be added to the
default cmdset before it becomes available.

"""

import re

from evennia import Command, DefaultCharacter
from evennia.utils import logger

# gender maps

_GENDER_PRONOUN_MAP = {
    "male": {"s": "he", "o": "him", "p": "his", "a": "his"},
    "female": {"s": "she", "o": "her", "p": "her", "a": "hers"},
    "neutral": {"s": "it", "o": "it", "p": "its", "a": "its"},
    "ambiguous": {"s": "they", "o": "them", "p": "their", "a": "theirs"},
}
_RE_GENDER_PRONOUN = re.compile(r"(?<!\|)\|(?!\|)[sSoOpPaA]")

# in-game command for setting the gender


class SetGender(Command):
    """
    Sets gender on yourself

    Usage:
      @gender male || female || neutral || ambiguous

    """

    key = "gender"
    aliases = "sex"
    locks = "call:all()"

    def func(self):
        """
        Implements the command.
        """
        caller = self.caller
        arg = self.args.strip().lower()
        if arg not in ("male", "female", "neutral", "ambiguous"):
            caller.msg("Usage: @gender male||female||neutral||ambiguous")
            return
        caller.db.gender = arg
        caller.msg("Your gender was set to %s." % arg)


# Gender-aware character class


class GenderCharacter(DefaultCharacter):
    """
    This is a Character class aware of gender.

    """

    def at_object_creation(self):
        """
        Called once when the object is created.
        """
        super().at_object_creation()
        self.db.gender = "ambiguous"

    def _get_pronoun(self, regex_match, source=None):
        """
        Get pronoun from the pronoun marker in the text. This is used as
        the callable for the re.sub function.

        Args:
            regex_match (MatchObject): the regular expression match.

        Notes:
            - `|s`, `|S`: Subjective form: he, she, it, He, She, It, They
            - `|o`, `|O`: Objective form: him, her, it, Him, Her, It, Them
            - `|p`, `|P`: Possessive form: his, her, its, His, Her, Its, Their
            - `|a`, `|A`: Absolute Possessive form: his, hers, its, His, Hers, Its, Theirs

        """
        if not source:
            source = self
        typ = regex_match.group()[1]  # "s", "O" etc
        gender = source.attributes.get("gender", default="ambiguous")
        gender = gender if gender in ("male", "female", "neutral") else "ambiguous"
        pronoun = _GENDER_PRONOUN_MAP[gender][typ.lower()]
        return pronoun.capitalize() if typ.isupper() else pronoun

    def msg(self, text=None, from_obj=None, session=None, **kwargs):
        """
        Emits something to a session attached to the object.
        Overloads the default msg() implementation to include
        gender-aware markers in output.

        Args:
            text (str or tuple, optional): The message to send. This
                is treated internally like any send-command, so its
                value can be a tuple if sending multiple arguments to
                the `text` oob command.
            from_obj (obj, optional): object that is sending. If
                given, at_msg_send will be called
            session (Session or list, optional): session or list of
                sessions to relay to, if any. If set, will
                force send regardless of MULTISESSION_MODE.
        Notes:
            `at_msg_receive` will be called on this Object.
            All extra kwargs will be passed on to the protocol.

        """
        if text is None:
            super().msg(from_obj=from_obj, session=session, **kwargs)
            return

        try:
            if text and isinstance(text, tuple):
                text = (
                    _RE_GENDER_PRONOUN.sub(
                        lambda x: self._get_pronoun(x, source=from_obj), text[0]
                    ),
                    *text[1:],
                )
            else:
                text = _RE_GENDER_PRONOUN.sub(lambda x: self._get_pronoun(x, source=from_obj), text)
        except TypeError:
            pass
        except Exception as e:
            logger.log_trace(e)

        super().msg(text, from_obj=from_obj, session=session, **kwargs)



"""
Start plugin services

This plugin module can define user-created services for the Portal to
start.

This module must handle all imports and setups required to start
twisted services (see examples in evennia.server.portal.portal). It
must also contain a function start_plugin_services(application).
Evennia will call this function with the main Portal application (so
your services can be added to it). The function should not return
anything. Plugin services are started last in the Portal startup
process.

"""


def start_plugin_services(portal):
    """
    This hook is called by Evennia, last in the Portal startup process.

    portal - a reference to the main portal application.
    """
    pass



"""
IDmapper extension to the default manager.
"""
from django.db.models.manager import Manager


class SharedMemoryManager(Manager):
    # TODO: improve on this implementation
    # We need a way to handle reverse lookups so that this model can
    # still use the singleton cache, but the active model isn't required
    # to be a SharedMemoryModel.
    def get(self, *args, **kwargs):
        """
        Data entity lookup.
        """
        items = list(kwargs)
        inst = None
        if len(items) == 1:
            # CL: support __exact
            key = items[0]
            if key.endswith("__exact"):
                key = key[: -len("__exact")]
            if key in ("pk", self.model._meta.pk.attname):
                try:
                    inst = self.model.get_cached_instance(kwargs[items[0]])
                    # we got the item from cache, but if this is a fk, check it's ours
                    if getattr(inst, str(self.field).split(".")[-1]) != self.instance:
                        inst = None
                except Exception:
                    pass
        if inst is None:
            inst = super().get(*args, **kwargs)
        return inst



r"""
Evennia settings file.

The available options are found in the default settings file found
here:

https://www.evennia.com/docs/latest/Setup/Settings-Default.html

Remember:

Don't copy more from the default file than you actually intend to
change; this will make sure that you don't overload upstream updates
unnecessarily.

When changing a setting requiring a file system path (like
path/to/actual/file.py), use GAME_DIR and EVENNIA_DIR to reference
your game folder and the Evennia library folders respectively. Python
paths (path.to.module) should be given relative to the game's root
folder (typeclasses.foo) whereas paths within the Evennia library
needs to be given explicitly (evennia.foo).

If you want to share your game dir, including its settings, you can
put secret game- or server-specific settings in secret_settings.py.

"""

# Use the defaults from Evennia unless explicitly overridden
from evennia.settings_default import *

######################################################################
# Evennia base server config
######################################################################

# This is the name of your game. Make it catchy!
SERVERNAME = {servername}


######################################################################
# Settings given in secret_settings.py override those in this file.
######################################################################
try:
    from server.conf.secret_settings import *
except ImportError:
    print("secret_settings.py file not found or failed to import.")



"""
Command handler

This module contains the infrastructure for accepting commands on the
command line. The processing of a command works as follows:

1. The calling object (caller) is analyzed based on its callertype.
2. Cmdsets are gathered from different sources:
   - object cmdsets: all objects at caller's location are scanned for non-empty
     cmdsets. This includes cmdsets on exits.
   - caller: the caller is searched for its own currently active cmdset.
   - account: lastly the cmdsets defined on caller.account are added.
3. The collected cmdsets are merged together to a combined, current cmdset.
4. If the input string is empty -> check for CMD_NOINPUT command in
   current cmdset or fallback to error message. Exit.
5. The Command Parser is triggered, using the current cmdset to analyze the
   input string for possible command matches.
6. If multiple matches are found -> check for CMD_MULTIMATCH in current
   cmdset, or fallback to error message. Exit.
7. If no match was found -> check for CMD_NOMATCH in current cmdset or
   fallback to error message. Exit.
8. At this point we have found a normal command. We assign useful variables to it that
   will be available to the command coder at run-time.
9. We have a unique cmdobject, primed for use. Call all hooks:
   `at_pre_cmd()`, `cmdobj.parse()`, `cmdobj.func()` and finally `at_post_cmd()`.
10. Return deferred that will fire with the return from `cmdobj.func()` (unused by default).

"""

import types
from collections import defaultdict
from copy import copy
from itertools import chain
from traceback import format_exc
from weakref import WeakValueDictionary

from django.conf import settings
from django.utils.translation import gettext as _
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.task import deferLater

from evennia.commands.cmdset import CmdSet
from evennia.commands.command import InterruptCommand
from evennia.utils import logger, utils
from evennia.utils.utils import string_suggestions

_IN_GAME_ERRORS = settings.IN_GAME_ERRORS

__all__ = ("cmdhandler", "InterruptCommand")
_GA = object.__getattribute__
_CMDSET_MERGE_CACHE = WeakValueDictionary()

# tracks recursive calls by each caller
# to avoid infinite loops (commands calling themselves)
_COMMAND_NESTING = defaultdict(lambda: 0)
_COMMAND_RECURSION_LIMIT = 10

# This decides which command parser is to be used.
# You have to restart the server for changes to take effect.
_COMMAND_PARSER = utils.variable_from_module(*settings.COMMAND_PARSER.rsplit(".", 1))

# System command names - import these variables rather than trying to
# remember the actual string constants. If not defined, Evennia
# hard-coded defaults are used instead.

# command to call if user just presses <return> with no input
CMD_NOINPUT = "__noinput_command"
# command to call if no command match was found
CMD_NOMATCH = "__nomatch_command"
# command to call if multiple command matches were found
CMD_MULTIMATCH = "__multimatch_command"
# command to call as the very first one when the user connects.
# (is expected to display the login screen)
CMD_LOGINSTART = "__unloggedin_look_command"


# Function for handling multiple command matches.
_SEARCH_AT_RESULT = utils.variable_from_module(*settings.SEARCH_AT_RESULT.rsplit(".", 1))

# Output strings. The first is the IN_GAME_ERRORS return, the second
# is the normal "production message to echo to the account.

_ERROR_UNTRAPPED = (
    _(
        """
An untrapped error occurred.
"""
    ),
    _(
        """
An untrapped error occurred. Please file a bug report detailing the steps to reproduce.
"""
    ),
)

_ERROR_CMDSETS = (
    _(
        """
A cmdset merger-error occurred. This is often due to a syntax
error in one of the cmdsets to merge.
"""
    ),
    _(
        """
A cmdset merger-error occurred. Please file a bug report detailing the
steps to reproduce.
"""
    ),
)

_ERROR_NOCMDSETS = (
    _(
        """
No command sets found! This is a critical bug that can have
multiple causes.
"""
    ),
    _(
        """
No command sets found! This is a sign of a critical bug.  If
disconnecting/reconnecting doesn't" solve the problem, try to contact
the server admin through" some other means for assistance.
"""
    ),
)

_ERROR_CMDHANDLER = (
    _(
        """
A command handler bug occurred. If this is not due to a local change,
please file a bug report with the Evennia project, including the
traceback and steps to reproduce.
"""
    ),
    _(
        """
A command handler bug occurred. Please notify staff - they should
likely file a bug report with the Evennia project.
"""
    ),
)

_ERROR_RECURSION_LIMIT = _(
    "Command recursion limit ({recursion_limit}) reached for '{raw_cmdname}' ({cmdclass})."
)


# delayed imports
_GET_INPUT = None


# helper functions
def err_helper(raw_string, cmdid=None):
    if cmdid is not None:
        return raw_string, {"cmdid": cmdid}
    return raw_string


def _msg_err(receiver, stringtuple, cmdid=None):
    """
    Helper function for returning an error to the caller.

    Args:
        receiver (Object): object to get the error message.
        stringtuple (tuple): tuple with two strings - one for the
            _IN_GAME_ERRORS mode (with the traceback) and one with the
            production string (with a timestamp) to be shown to the user.

    """
    string = _("{traceback}\n{errmsg}\n(Traceback was logged {timestamp}).")
    timestamp = logger.timeformat()
    tracestring = format_exc()
    logger.log_trace()
    if _IN_GAME_ERRORS:
        out = string.format(
            traceback=tracestring, errmsg=stringtuple[0].strip(), timestamp=timestamp
        ).strip()
    else:
        out = string.format(
            traceback=tracestring.splitlines()[-1],
            errmsg=stringtuple[1].strip(),
            timestamp=timestamp,
        ).strip()
    receiver.msg(err_helper(out, cmdid=cmdid))


def _process_input(caller, prompt, result, cmd, generator):
    """
    Specifically handle the get_input value to send to _progressive_cmd_run as
    part of yielding from a Command's `func`.

    Args:
        caller (Character, Account or Session): the caller.
        prompt (str): The sent prompt.
        result (str): The unprocessed answer.
        cmd (Command): The command itself.
        generator (GeneratorType): The generator.

    Returns:
        result (bool): Always `False` (stop processing).

    """
    # We call it using a Twisted deferLater to make sure the input is properly closed.
    deferLater(reactor, 0, _progressive_cmd_run, cmd, generator, response=result)
    return False


def _progressive_cmd_run(cmd, generator, response=None):
    """
    Progressively call the command that was given in argument. Used
    when `yield` is present in the Command's `func()` method.

    Args:
        cmd (Command): the command itself.
        generator (GeneratorType): the generator describing the processing.
        reponse (str, optional): the response to send to the generator.

    Raises:
        ValueError: If the func call yields something not identifiable as a
            time-delay or a string prompt.

    Note:
        This function is responsible for executing the command, if
        the func() method contains 'yield' instructions.  The yielded
        value will be accessible at each step and will affect the
        process.  If the value is a number, just delay the execution
        of the command.  If it's a string, wait for the user input.

    """
    global _GET_INPUT
    if not _GET_INPUT:
        from evennia.utils.evmenu import get_input as _GET_INPUT

    try:
        if response is None:
            value = next(generator)
        else:
            value = generator.send(response)
    except StopIteration:
        # duplicated from cmdhandler._run_command, to have these
        # run in the right order while staying inside the deferred
        cmd.at_post_cmd()
        if cmd.save_for_next:
            # store a reference to this command, possibly
            # accessible by the next command.
            cmd.caller.ndb.last_cmd = copy(cmd)
        else:
            cmd.caller.ndb.last_cmd = None
    else:
        if isinstance(value, (int, float)):
            utils.delay(value, _progressive_cmd_run, cmd, generator)
        elif isinstance(value, str):
            _GET_INPUT(cmd.caller, value, _process_input, cmd=cmd, generator=generator)
        else:
            raise ValueError("unknown type for a yielded value in command: {}".format(type(value)))


# custom Exceptions


class NoCmdSets(Exception):
    "No cmdsets found. Critical error."
    pass


class ExecSystemCommand(Exception):
    "Run a system command"

    def __init__(self, syscmd, sysarg):
        self.args = (syscmd, sysarg)  # needed by exception error handling
        self.syscmd = syscmd
        self.sysarg = sysarg


class ErrorReported(Exception):
    "Re-raised when a subsructure already reported the error"

    def __init__(self, raw_string):
        self.args = (raw_string,)
        self.raw_string = raw_string


# Helper function
def generate_cmdset_providers(called_by, session=None):
    cmdset_providers = dict()
    cmdset_providers.update(called_by.get_cmdset_providers())
    if session and session is not called_by:
        cmdset_providers.update(session.get_cmdset_providers())

    cmdset_providers_list = list(cmdset_providers.values())
    cmdset_providers_list.sort(key=lambda x: getattr(x, "cmdset_provider_order", 0))
    # sort the dictionary by priority. This can be done because Python now cares about dictionary insert order.
    cmdset_providers = {c.cmdset_provider_type: c for c in cmdset_providers_list}

    if not cmdset_providers:
        raise RuntimeError("cmdhandler: no command objects found.")

    # the caller will be the one to receive messages and excert its permissions.
    # we assign the caller with preference 'bottom up'
    caller = cmdset_providers_list[-1]

    cmdset_providers_errors_list = sorted(
        cmdset_providers_list, key=lambda x: getattr(x, "cmdset_provider_error_order", 0)
    )

    # The error_to is the default recipient for errors. Tries to make sure an account
    # does not get spammed for errors while preserving character mirroring.
    error_to = cmdset_providers_errors_list[-1]

    return cmdset_providers, cmdset_providers_list, cmdset_providers_errors_list, caller, error_to


@inlineCallbacks
def get_and_merge_cmdsets(
    caller, cmdset_providers, callertype, raw_string, report_to=None, cmdid=None
):
    """
    Gather all relevant cmdsets and merge them.

    Args:
        caller (Session, Account or Object): The entity executing the command. Which
            type of object this is depends on the current game state; for example
            when the user is not logged in, this will be a Session, when being OOC
            it will be an Account and when puppeting an object this will (often) be
            a Character Object. In the end it depends on where the cmdset is stored.
        cmdset_providers (list): A list of sorted objects which provide cmdsets.
        callertype (str): This identifies caller as either "account", "object" or "session"
            to avoid having to do this check internally.
        raw_string (str): The input string. This is only used for error reporting.
        report_to (Object, optional): If given, this object will receive error messages

    Returns:
        cmdset (Deferred): This deferred fires with the merged cmdset
        result once merger finishes.

    Notes:
        The cdmsets are merged in order or generality, so that the
        Object's cmdset is merged last (and will thus take precedence
        over same-named and same-prio commands on Account and Session).

    """
    try:

        @inlineCallbacks
        def _get_local_obj_cmdsets(obj):
            """
            Helper-method; Get Object-level cmdsets

            """
            # Gather cmdsets from location, objects in location or carried
            try:
                local_obj_cmdsets = [None]
                try:
                    location = obj.location
                except Exception:
                    location = None
                if location:
                    # Gather all cmdsets stored on objects in the room and
                    # also in the caller's inventory and the location itself
                    local_objlist = yield (
                        location.contents_get(exclude=obj) + obj.contents_get() + [location]
                    )
                    local_objlist = [o for o in local_objlist if not o._is_deleted]
                    for lobj in local_objlist:
                        try:
                            # call hook in case we need to do dynamic changing to cmdset
                            _GA(lobj, "at_cmdset_get")(caller=caller)
                        except Exception:
                            logger.log_trace()
                    # the call-type lock is checked here, it makes sure an account
                    # is not seeing e.g. the commands on a fellow account (which is why
                    # the no_superuser_bypass must be True)
                    local_obj_cmdsets = yield list(
                        chain.from_iterable(
                            lobj.cmdset.cmdset_stack
                            for lobj in local_objlist
                            if (
                                lobj.cmdset.current
                                and lobj.access(
                                    caller, access_type="call", no_superuser_bypass=True
                                )
                            )
                        )
                    )
                    for cset in local_obj_cmdsets:
                        # This is necessary for object sets, or we won't be able to
                        # separate the command sets from each other in a busy room. We
                        # only keep the setting if duplicates were set to False/True
                        # explicitly.
                        cset.old_duplicates = cset.duplicates
                        cset.duplicates = True if cset.duplicates is None else cset.duplicates
                returnValue(local_obj_cmdsets)
            except Exception:
                _msg_err(caller, _ERROR_CMDSETS)
                raise ErrorReported(raw_string)

        @inlineCallbacks
        def _get_cmdsets(obj, current):
            """
            Helper method; Get cmdset while making sure to trigger all
            hooks safely. Returns the stack and the valid options.

            """
            try:
                yield obj.at_cmdset_get(caller=caller, current=current)
            except Exception:
                _msg_err(caller, _ERROR_CMDSETS)
                raise ErrorReported(raw_string)
            try:
                returnValue(obj.get_cmdsets(caller=caller, current=current))
            except AttributeError:
                returnValue((CmdSet(), []))

        local_obj_cmdsets = []

        current_cmdset = CmdSet()
        object_cmdsets = list()
        for cmdobj in cmdset_providers:
            current, cur_cmdsets = yield _get_cmdsets(cmdobj, current_cmdset)
            if current:
                current_cmdset = current_cmdset + current
            if cur_cmdsets:
                object_cmdsets += cur_cmdsets
            match cmdobj.cmdset_provider_type:
                case "object":
                    if not current.no_objs:
                        local_obj_cmdsets = yield _get_local_obj_cmdsets(cmdobj)
                        if current.no_exits:
                            # filter out all exits
                            local_obj_cmdsets = [
                                cmdset for cmdset in local_obj_cmdsets if cmdset.key != "ExitCmdSet"
                            ]
                        object_cmdsets += local_obj_cmdsets

        # weed out all non-found sets
        cmdsets = yield [
            cmdset for cmdset in object_cmdsets if cmdset and cmdset.key != "_EMPTY_CMDSET"
        ]
        # report cmdset errors to user (these should already have been logged)
        yield [
            report_to.msg(err_helper(cmdset.errmessage, cmdid=cmdid))
            for cmdset in cmdsets
            if cmdset.key == "_CMDSET_ERROR"
        ]

        if cmdsets:
            # faster to do tuple on list than to build tuple directly
            mergehash = tuple([id(cmdset) for cmdset in cmdsets])
            if mergehash in _CMDSET_MERGE_CACHE:
                # cached merge exist; use that
                cmdset = _CMDSET_MERGE_CACHE[mergehash]
            else:
                # we group and merge all same-prio cmdsets separately (this avoids
                # order-dependent clashes in certain cases, such as
                # when duplicates=True)
                tempmergers = {}
                for cmdset in cmdsets:
                    prio = cmdset.priority
                    if prio in tempmergers:
                        # merge same-prio cmdset together separately
                        tempmergers[prio] = yield tempmergers[prio] + cmdset
                    else:
                        tempmergers[prio] = cmdset

                # sort cmdsets after reverse priority (highest prio are merged in last)
                sorted_cmdsets = yield sorted(list(tempmergers.values()), key=lambda x: x.priority)

                # Merge all command sets into one, beginning with the lowest-prio one
                cmdset = sorted_cmdsets[0]
                for merging_cmdset in sorted_cmdsets[1:]:
                    cmdset = yield cmdset + merging_cmdset
                # store the original, ungrouped set for diagnosis
                cmdset.merged_from = cmdsets
                # cache
                _CMDSET_MERGE_CACHE[mergehash] = cmdset
        else:
            cmdset = None
        for cset in (cset for cset in local_obj_cmdsets if cset):
            cset.duplicates = cset.old_duplicates
        # important - this syncs the CmdSetHandler's .current field with the
        # true current cmdset!
        # TODO - removed because this causes cmdset overlaps across sessions/accounts
        # - see https://github.com/evennia/evennia/issues/2855
        # if cmdset:
        #     caller.cmdset.current = cmdset

        returnValue(cmdset)
    except ErrorReported:
        raise
    except Exception:
        _msg_err(caller, _ERROR_CMDSETS)
        raise
        # raise ErrorReported


# Main command-handler function


@inlineCallbacks
def cmdhandler(
    called_by,
    raw_string,
    _testing=False,
    callertype="session",
    session=None,
    cmdobj=None,
    cmdobj_key=None,
    **kwargs,
):
    """
    This is the main mechanism that handles any string sent to the engine.

    Args:
        called_by (Session, Account or Object): Object from which this
            command was called. which this was called from.  What this is
            depends on the game state.
        raw_string (str): The command string as given on the command line.
        _testing (bool, optional): Used for debug purposes and decides if we
            should actually execute the command or not. If True, the
            command instance will be returned.
        callertype (str, optional): One of "session", "account" or
            "object". These are treated in decending order, so when the
            Session is the caller, it will merge its own cmdset into
            cmdsets from both Account and eventual puppeted Object (and
            cmdsets in its room etc). An Account will only include its own
            cmdset and the Objects and so on. Merge order is the same
            order, so that Object cmdsets are merged in last, giving them
            precendence for same-name and same-prio commands.
        session (Session, optional): Relevant if callertype is "account" - the session will help
            retrieve the correct cmdsets from puppeted objects.
        cmdobj (Command, optional): If given a command instance, this will be executed using
            `called_by` as the caller, `raw_string` representing its arguments and (optionally)
            `cmdobj_key` as its input command name. No cmdset lookup will be performed but
            all other options apply as normal. This allows for running a specific Command
            within the command system mechanism.
        cmdobj_key (string, optional): Used together with `cmdobj` keyword to specify
            which cmdname should be assigned when calling the specified Command instance. This
            is made available as `self.cmdstring` when the Command runs.
            If not given, the command will be assumed to be called as `cmdobj.key`.

    Keyword Args:
        kwargs (any): other keyword arguments will be assigned as named variables on the
            retrieved command object *before* it is executed. This is unused
            in default Evennia but may be used by code to set custom flags or
            special operating conditions for a command as it executes.

    Returns:
        deferred (Deferred): This deferred is fired with the return
        value of the command's `func` method.  This is not used in
        default Evennia.

    """
    cmdid = kwargs.get("cmdid", None)

    @inlineCallbacks
    def _run_command(cmd, cmdname, args, raw_cmdname, cmdset, session, account, cmdset_providers):
        """
        Helper function: This initializes and runs the Command
        instance once the parser has identified it as either a normal
        command or one of the system commands.

        Args:
            cmd (Command): Command object
            cmdname (str): Name of command
            args (str): extra text entered after the identified command
            raw_cmdname (str): Name of Command, unaffected by eventual
                prefix-stripping (if no prefix-stripping, this is the same
                as cmdname).
            cmdset (CmdSet): Command sert the command belongs to (if any)..
            session (Session): Session of caller (if any).
            account (Account): Account of caller (if any).
            cmdset_providers (dict): Dictionary of all cmdset-providing objects.

        Returns:
            deferred (Deferred): this will fire with the return of the
                command's `func` method.

        Raises:
            RuntimeError: If command recursion limit was reached.

        """
        global _COMMAND_NESTING
        try:
            # Assign useful variables to the instance
            cmd.caller = caller
            cmd.cmdname = cmdname
            cmd.raw_cmdname = raw_cmdname
            cmd.cmdstring = cmdname  # deprecated
            cmd.args = args
            cmd.cmdset = cmdset
            cmd.cmdset_providers = cmdset_providers.copy()
            cmd.session = session
            cmd.account = account
            cmd.raw_string = unformatted_raw_string
            # cmd.obj  # set via on-object cmdset handler for each command,
            # since this may be different for every command when
            # merging multiple cmdsets

            if _testing:
                # only return the command instance
                returnValue(cmd)

            # assign custom kwargs to found cmd object
            for key, val in kwargs.items():
                setattr(cmd, key, val)

            _COMMAND_NESTING[called_by] += 1
            if _COMMAND_NESTING[called_by] > _COMMAND_RECURSION_LIMIT:
                err = _ERROR_RECURSION_LIMIT.format(
                    recursion_limit=_COMMAND_RECURSION_LIMIT,
                    raw_cmdname=raw_cmdname,
                    cmdclass=cmd.__class__,
                )
                raise RuntimeError(err)

            # pre-command hook
            abort = yield cmd.at_pre_cmd()
            if abort:
                # abort sequence
                returnValue(abort)

            # Parse and execute
            yield cmd.parse()

            # main command code
            # (return value is normally None)
            ret = cmd.func()
            if isinstance(ret, types.GeneratorType):
                # cmd.func() is a generator, execute progressively
                _progressive_cmd_run(cmd, ret)
                ret = yield ret
                # note that the _progressive_cmd_run will itself run
                # the at_post_cmd etc as it finishes; this is a bit of
                # code duplication but there seems to be no way to
                # catch the StopIteration here (it's not in the same
                # frame since this is in a deferred chain)
            else:
                ret = yield ret
                # post-command hook
                yield cmd.at_post_cmd()

                if cmd.save_for_next:
                    # store a reference to this command, possibly
                    # accessible by the next command.
                    caller.ndb.last_cmd = yield copy(cmd)
                else:
                    caller.ndb.last_cmd = None

            # return result to the deferred
            returnValue(ret)

        except InterruptCommand:
            # Do nothing, clean exit
            pass
        except Exception:
            _msg_err(caller, _ERROR_UNTRAPPED)
            raise ErrorReported(raw_string)
        finally:
            _COMMAND_NESTING[called_by] -= 1

    (
        cmdset_providers,
        cmdset_providers_list,
        cmdset_providers_list_error,
        caller,
        error_to,
    ) = generate_cmdset_providers(called_by, session=session)

    account = cmdset_providers.get("account", None)

    try:  # catch bugs in cmdhandler itself
        try:  # catch special-type commands
            if cmdobj:
                # the command object is already given
                cmd = cmdobj() if callable(cmdobj) else cmdobj
                cmdname = cmdobj_key if cmdobj_key else cmd.key
                args = raw_string
                unformatted_raw_string = "%s%s" % (cmdname, args)
                cmdset = None
                raw_cmdname = cmdname
                # session = session
                # account = account

            else:
                # no explicit cmdobject given, figure it out
                cmdset = yield get_and_merge_cmdsets(
                    caller, cmdset_providers_list, callertype, raw_string, cmdid=cmdid
                )
                if not cmdset:
                    # this is bad and shouldn't happen.
                    raise NoCmdSets
                # store the completely unmodified raw string - including
                # whitespace and eventual prefixes-to-be-stripped.
                unformatted_raw_string = raw_string
                raw_string = raw_string.strip()
                if not raw_string:
                    # Empty input. Test for system command instead.
                    syscmd = yield cmdset.get(CMD_NOINPUT)
                    sysarg = ""
                    raise ExecSystemCommand(syscmd, sysarg)
                # Parse the input string and match to available cmdset.
                # This also checks for permissions, so all commands in match
                # are commands the caller is allowed to call.
                matches = yield _COMMAND_PARSER(raw_string, cmdset, caller)

                # Deal with matches

                if len(matches) > 1:
                    # We have a multiple-match
                    syscmd = yield cmdset.get(CMD_MULTIMATCH)
                    sysarg = _("There were multiple matches.")
                    if syscmd:
                        # use custom CMD_MULTIMATCH
                        syscmd.matches = matches
                    else:
                        # fall back to default error handling
                        sysarg = yield _SEARCH_AT_RESULT(
                            [match[2] for match in matches], caller, query=matches[0][0]
                        )
                    raise ExecSystemCommand(syscmd, sysarg)

                cmdname, args, cmd, raw_cmdname = "", "", None, ""
                if len(matches) == 1:
                    # We have a unique command match. But it may still be invalid.
                    match = matches[0]
                    cmdname, args, cmd, raw_cmdname = (match[0], match[1], match[2], match[5])

                if not matches:
                    # No commands match our entered command
                    syscmd = yield cmdset.get(CMD_NOMATCH)
                    if syscmd:
                        # use custom CMD_NOMATCH command
                        sysarg = raw_string
                    else:
                        # fallback to default error text
                        sysarg = _("Command '{command}' is not available.").format(
                            command=raw_string
                        )
                        suggestions = string_suggestions(
                            raw_string,
                            cmdset.get_all_cmd_keys_and_aliases(caller),
                            cutoff=0.7,
                            maxnum=3,
                        )
                        if suggestions:
                            sysarg += _(" Maybe you meant {command}?").format(
                                command=utils.list_to_string(
                                    suggestions, endsep=_("or"), addquote=True
                                )
                            )
                        else:
                            sysarg += _(' Type "help" for help.')
                    raise ExecSystemCommand(syscmd, sysarg)

            if not cmd.retain_instance:
                # making a copy allows multiple users to share the command also when yield is used
                cmd = copy(cmd)

            # A normal command.
            ret = yield _run_command(
                cmd, cmdname, args, raw_cmdname, cmdset, session, account, cmdset_providers
            )
            returnValue(ret)

        except ErrorReported as exc:
            # this error was already reported, so we
            # catch it here and don't pass it on.
            logger.log_err("User input was: '%s'." % exc.raw_string)

        except ExecSystemCommand as exc:
            # Not a normal command: run a system command, if available,
            # or fall back to a return string.
            syscmd = exc.syscmd
            sysarg = exc.sysarg

            if syscmd:
                ret = yield _run_command(
                    syscmd,
                    syscmd.key,
                    sysarg,
                    unformatted_raw_string,
                    cmdset,
                    session,
                    account,
                    cmdset_providers,
                )
                returnValue(ret)
            elif sysarg:
                # return system arg
                error_to.msg(err_helper(exc.sysarg, cmdid=cmdid))

        except NoCmdSets:
            # Critical error.
            logger.log_err("No cmdsets found: %s" % caller)
            error_to.msg(err_helper(_ERROR_NOCMDSETS, cmdid=cmdid))

        except Exception:
            # We should not end up here. If we do, it's a programming bug.
            _msg_err(error_to, _ERROR_UNTRAPPED)

    except Exception:
        # This catches exceptions in cmdhandler exceptions themselves
        _msg_err(error_to, _ERROR_CMDHANDLER)



"""
Helper utils for admin views.

"""

import importlib

from evennia.utils.utils import get_all_cmdsets, get_all_typeclasses, inherits_from


def get_and_load_typeclasses(parent=None, excluded_parents=None):
    """
    Get all typeclasses. We we need to initialize things here
    for them to be actually available in the admin process.
    This is intended to be used with forms.ChoiceField.

    Args:
        parent (str or class, optional): Limit selection to this class and its children
            (at any distance).
        exclude (list): Class-parents to exclude from the resulting list. All
            children of these paretns will be skipped.

    Returns:
        list: A list of (str, str), the way ChoiceField wants it.

    """
    # this is necessary in order to have typeclasses imported and accessible
    # in the inheritance tree.
    import evennia

    evennia._init()

    # this return a dict (path: class}
    tmap = get_all_typeclasses(parent=parent)

    # filter out any excludes
    excluded_parents = excluded_parents or []
    tpaths = [
        path
        for path, tclass in tmap.items()
        if not any(inherits_from(tclass, excl) for excl in excluded_parents)
    ]

    # sort so we get custom paths (not in evennia repo) first
    tpaths = sorted(tpaths, key=lambda k: (1 if k.startswith("evennia.") else 0, k))

    # the base models are not typeclasses so we filter them out
    tpaths = [
        path
        for path in tpaths
        if path
        not in (
            "evennia.objects.models.ObjectDB",
            "evennia.accounts.models.AccountDB",
            "evennia.scripts.models.ScriptDB",
            "evennia.comms.models.ChannelDB",
        )
    ]

    # return on form excepted by ChoiceField
    return [(path, path) for path in tpaths if path]


def get_and_load_cmdsets(parent=None, excluded_parents=None):
    """
    Get all cmdsets available or as children based on a parent cmdset.  We need
    to initialize things here to make sure as much as possible is loaded in the
    admin process. This is intended to be used with forms.ChoiceField.

    Args:
        parent (str, optional): Python-path to the parent cmdset, if any.
        excluded_parents (list): A list of cmset-paths to exclude from the result.

    Returns:
        list: A list of (str, str), the way ChoiceField wants it.

    """
    # we must do this to have cmdsets imported and accessible in the inheritance tree.
    import evennia

    evennia._init()

    cmap = get_all_cmdsets(parent)

    excluded_parents = excluded_parents or []
    cpaths = [path for path in cmap if not any(path == excluded for excluded in excluded_parents)]

    cpaths = sorted(cpaths, key=lambda k: (1 if k.startswith("evennia.") else 0, k))

    # return on form expected by ChoiceField
    return [("", "-")] + [(path, path) for path in cpaths if path]



"""
Clothing - Provides a typeclass and commands for wearable clothing,
which is appended to a character's description when worn.

Evennia contribution - Tim Ashley Jenkins 2017

Clothing items, when worn, are added to the character's description
in a list. For example, if wearing the following clothing items:

    a thin and delicate necklace
    a pair of regular ol' shoes
    one nice hat
    a very pretty dress

A character's description may look like this:

    Superuser(#1)
    This is User #1.

    Superuser is wearing one nice hat, a thin and delicate necklace,
    a very pretty dress and a pair of regular ol' shoes.

Characters can also specify the style of wear for their clothing - I.E.
to wear a scarf 'tied into a tight knot around the neck' or 'draped
loosely across the shoulders' - to add an easy avenue of customization.
For example, after entering:

    wear scarf draped loosely across the shoulders

The garment appears like so in the description:

    Superuser(#1)
    This is User #1.

    Superuser is wearing a fanciful-looking scarf draped loosely
    across the shoulders.

Items of clothing can be used to cover other items, and many options
are provided to define your own clothing types and their limits and
behaviors. For example, to have undergarments automatically covered
by outerwear, or to put a limit on the number of each type of item
that can be worn. The system as-is is fairly freeform - you
can cover any garment with almost any other, for example - but it
can easily be made more restrictive, and can even be tied into a
system for armor or other equipment.

To install, import this module and have your default character
inherit from ClothedCharacter in your game's characters.py file:

    from evennia.contrib.game_systems.clothing import ClothedCharacter

    class Character(ClothedCharacter):

And then add ClothedCharacterCmdSet in your character set in your
game's commands/default_cmdsets.py:

    from evennia.contrib.game_systems.clothing import ClothedCharacterCmdSet

    class CharacterCmdSet(default_cmds.CharacterCmdSet):
         ...
         at_cmdset_creation(self):

             super().at_cmdset_creation()
             ...
             self.add(ClothedCharacterCmdSet)    # <-- add this

From here, you can use the default builder commands to create clothes
with which to test the system:

    @create a pretty shirt : evennia.contrib.game_systems.clothing.ContribClothing
    @set shirt/clothing_type = 'top'
    wear shirt

"""
from collections import defaultdict

from django.conf import settings

from evennia import DefaultCharacter, DefaultObject, default_cmds
from evennia.commands.default.muxcommand import MuxCommand
from evennia.utils import at_search_result, evtable, inherits_from, iter_to_str

# Options start here.
# Maximum character length of 'wear style' strings, or None for unlimited.
WEARSTYLE_MAXLENGTH = getattr(settings, "CLOTHING_WEARSTYLE_MAXLENGTH", 50)

# The rest of these options have to do with clothing types. ContribClothing types are optional,
# but can be used to give better control over how different items of clothing behave. You
# can freely add, remove, or change clothing types to suit the needs of your game and use
# the options below to affect their behavior.

# The order in which clothing types appear on the description. Untyped clothing or clothing
# with a type not given in this list goes last.
CLOTHING_TYPE_ORDER = getattr(
    settings,
    "CLOTHING_TYPE_ORDERED",
    [
        "hat",
        "jewelry",
        "top",
        "undershirt",
        "gloves",
        "fullbody",
        "bottom",
        "underpants",
        "socks",
        "shoes",
        "accessory",
    ],
)
# The maximum number of each type of clothes that can be worn. Unlimited if untyped or not specified.
CLOTHING_TYPE_LIMIT = getattr(
    settings, "CLOTHING_TYPE_LIMIT", {"hat": 1, "gloves": 1, "socks": 1, "shoes": 1}
)
# The maximum number of clothing items that can be worn, or None for unlimited.
CLOTHING_OVERALL_LIMIT = getattr(settings, "CLOTHING_OVERALL_LIMIT", 20)
# What types of clothes will automatically cover what other types of clothes when worn.
# Note that clothing only gets auto-covered if it's already worn when you put something
# on that auto-covers it - for example, it's perfectly possible to have your underpants
# showing if you put them on after your pants!
CLOTHING_TYPE_AUTOCOVER = getattr(
    settings,
    "CLOTHING_TYPE_AUTOCOVER",
    {
        "top": ["undershirt"],
        "bottom": ["underpants"],
        "fullbody": ["undershirt", "underpants"],
        "shoes": ["socks"],
    },
)
# Types of clothes that can't be used to cover other clothes.
CLOTHING_TYPE_CANT_COVER_WITH = getattr(settings, "CLOTHING_TYPE_AUTOCOVER", ["jewelry"])


# HELPER FUNCTIONS START HERE
def order_clothes_list(clothes_list):
    """
    Orders a given clothes list by the order specified in CLOTHING_TYPE_ORDER.

    Args:
        clothes_list (list): List of clothing items to put in order

    Returns:
        ordered_clothes_list (list): The same list as passed, but re-ordered
                                     according to the hierarchy of clothing types
                                     specified in CLOTHING_TYPE_ORDER.
    """
    ordered_clothes_list = clothes_list
    # For each type of clothing that exists...
    for current_type in reversed(CLOTHING_TYPE_ORDER):
        # Check each item in the given clothes list.
        for clothes in clothes_list:
            # If the item has a clothing type...
            if clothes.db.clothing_type:
                item_type = clothes.db.clothing_type
                # And the clothing type matches the current type...
                if item_type == current_type:
                    # Move it to the front of the list!
                    ordered_clothes_list.remove(clothes)
                    ordered_clothes_list.insert(0, clothes)
    return ordered_clothes_list


def get_worn_clothes(character, exclude_covered=False):
    """
    Get a list of clothes worn by a given character.

    Args:
        character (obj): The character to get a list of worn clothes from.

    Keyword Args:
        exclude_covered (bool): If True, excludes clothes covered by other
                                clothing from the returned list.

    Returns:
        ordered_clothes_list (list): A list of clothing items worn by the
                                     given character, ordered according to
                                     the CLOTHING_TYPE_ORDER option specified
                                     in this module.
    """
    clothes_list = []
    for thing in character.contents:
        # If uncovered or not excluding covered items
        if not thing.db.covered_by or exclude_covered is False:
            # If 'worn' is True, add to the list
            if thing.db.worn:
                clothes_list.append(thing)
    # Might as well put them in order here too.
    ordered_clothes_list = order_clothes_list(clothes_list)
    return ordered_clothes_list


def clothing_type_count(clothes_list):
    """
    Returns a dictionary of the number of each clothing type
    in a given list of clothing objects.

    Args:
        clothes_list (list): A list of clothing items from which
                             to count the number of clothing types
                             represented among them.

    Returns:
        types_count (dict): A dictionary of clothing types represented
                            in the given list and the number of each
                            clothing type represented.
    """
    types_count = {}
    for garment in clothes_list:
        if garment.db.clothing_type:
            type = garment.db.clothing_type
            if type not in list(types_count.keys()):
                types_count[type] = 1
            else:
                types_count[type] += 1
    return types_count


def single_type_count(clothes_list, type):
    """
    Returns an integer value of the number of a given type of clothing in a list.

    Args:
        clothes_list (list): List of clothing objects to count from
        type (str): Clothing type to count

    Returns:
        type_count (int): Number of garments of the specified type in the given
                          list of clothing objects
    """
    type_count = 0
    for garment in clothes_list:
        if garment.db.clothing_type:
            if garment.db.clothing_type == type:
                type_count += 1
    return type_count


class ContribClothing(DefaultObject):
    def wear(self, wearer, wearstyle, quiet=False):
        """
        Sets clothes to 'worn' and optionally echoes to the room.

        Args:
            wearer (obj): character object wearing this clothing object
            wearstyle (True or str): string describing the style of wear or True for none

        Keyword Args:
            quiet (bool): If false, does not message the room

        Notes:
            Optionally sets db.worn with a 'wearstyle' that appends a short passage to
            the end of the name  of the clothing to describe how it's worn that shows
            up in the wearer's desc - I.E. 'around his neck' or 'tied loosely around
            her waist'. If db.worn is set to 'True' then just the name will be shown.
        """
        # Set clothing as worn
        self.db.worn = wearstyle
        # Auto-cover appropriate clothing types
        to_cover = []
        if clothing_type := self.db.clothing_type:
            if autocover_types := CLOTHING_TYPE_AUTOCOVER.get(clothing_type):
                to_cover.extend(
                    [
                        garment
                        for garment in get_worn_clothes(wearer)
                        if garment.db.clothing_type in autocover_types
                    ]
                )
        for garment in to_cover:
            garment.db.covered_by = self

        # Echo a message to the room
        if not quiet:
            if type(wearstyle) is str:
                message = f"$You() $conj(wear) {self.name} {wearstyle}"
            else:
                message = f"$You() $conj(put) on {self.name}"
            if to_cover:
                message += f", covering {iter_to_str(to_cover)}"
            wearer.location.msg_contents(message + ".", from_obj=wearer)

    def remove(self, wearer, quiet=False):
        """
        Removes worn clothes and optionally echoes to the room.

        Args:
            wearer (obj): character object wearing this clothing object

        Keyword Args:
            quiet (bool): If false, does not message the room
        """
        self.db.worn = False
        uncovered_list = []

        # Check to see if any other clothes are covered by this object.
        for thing in wearer.contents:
            if thing.db.covered_by == self:
                thing.db.covered_by = False
                uncovered_list.append(thing.name)
        # Echo a message to the room
        if not quiet:
            remove_message = f"$You() $conj(remove) {self.name}"
            if len(uncovered_list) > 0:
                remove_message += f", revealing {iter_to_str(uncovered_list)}"
            wearer.location.msg_contents(remove_message + ".", from_obj=wearer)

    def at_get(self, getter):
        """
        Makes absolutely sure clothes aren't already set as 'worn'
        when they're picked up, in case they've somehow had their
        location changed without getting removed.
        """
        self.db.worn = False

    def at_pre_move(self, destination, **kwargs):
        """
        Called just before starting to move this object to
        destination. Return False to abort move.

        Notes:
            If this method returns False/None, the move is cancelled
            before it is even started.
        """
        # Covered clothing cannot be removed, dropped, or otherwise relocated
        if self.db.covered_by:
            return False
        return True


class ClothedCharacter(DefaultCharacter):
    """
    Character that displays worn clothing when looked at. You can also
    just copy the return_appearance hook defined below to your own game's
    character typeclass.
    """

    def get_display_desc(self, looker, **kwargs):
        """
        Get the 'desc' component of the object description. Called by `return_appearance`.

        Args:
            looker (Object): Object doing the looking.
            **kwargs: Arbitrary data for use when overriding.
        Returns:
            str: The desc display string.
        """
        desc = self.db.desc

        outfit_list = []
        # Append worn, uncovered clothing to the description
        for garment in get_worn_clothes(self, exclude_covered=True):
            wearstyle = garment.db.worn
            if type(wearstyle) is str:
                outfit_list.append(f"{garment.name} {wearstyle}")
            else:
                outfit_list.append(garment.name)

        # Create outfit string
        if outfit_list:
            outfit = (
                f"{self.get_display_name(looker, **kwargs)} is wearing {iter_to_str(outfit_list)}."
            )
        else:
            outfit = f"{self.get_display_name(looker, **kwargs)} is wearing nothing."

        # Add on to base description
        if desc:
            desc += f"\n\n{outfit}"
        else:
            desc = outfit

        return desc

    def get_display_things(self, looker, **kwargs):
        """
        Get the 'things' component of the object's contents. Called by `return_appearance`.

        Args:
            looker (Object): Object doing the looking.
            **kwargs: Arbitrary data for use when overriding.
        Returns:
            str: A string describing the things in object.
        """

        def _filter_visible(obj_list):
            return (
                obj
                for obj in obj_list
                if obj != looker and obj.access(looker, "view") and not obj.db.worn
            )

        # sort and handle same-named things
        things = _filter_visible(self.contents_get(content_type="object"))

        grouped_things = defaultdict(list)
        for thing in things:
            grouped_things[thing.get_display_name(looker, **kwargs)].append(thing)

        thing_names = []
        for thingname, thinglist in sorted(grouped_things.items()):
            nthings = len(thinglist)
            thing = thinglist[0]
            singular, plural = thing.get_numbered_name(nthings, looker, key=thingname)
            thing_names.append(singular if nthings == 1 else plural)
        thing_names = iter_to_str(thing_names)
        return (
            f"\n{self.get_display_name(looker, **kwargs)} is carrying {thing_names}"
            if thing_names
            else ""
        )


# COMMANDS START HERE


class CmdWear(MuxCommand):
    """
    Puts on an item of clothing you are holding.

    Usage:
      wear <obj> [=] [wear style]

    Examples:
      wear red shirt
      wear scarf wrapped loosely about the shoulders
      wear blue hat = at a jaunty angle

    All the clothes you are wearing are appended to your description.
    If you provide a 'wear style' after the command, the message you
    provide will be displayed after the clothing's name.
    """

    key = "wear"
    help_category = "clothing"

    def func(self):
        if not self.args:
            self.caller.msg("Usage: wear <obj> [=] [wear style]")
            return
        if not self.rhs:
            # check if the whole string is an object
            clothing = self.caller.search(self.lhs, candidates=self.caller.contents, quiet=True)
            if not clothing:
                # split out the first word as the object and the rest as the wearstyle
                argslist = self.lhs.split()
                self.lhs = argslist[0]
                self.rhs = " ".join(argslist[1:])
                clothing = self.caller.search(self.lhs, candidates=self.caller.contents)
            else:
                # pass the result through the search-result hook
                clothing = at_search_result(clothing, self.caller, self.lhs)

        else:
            # it had an explicit separator - just do a normal search for the lhs
            clothing = self.caller.search(self.lhs, candidates=self.caller.contents)

        if not clothing:
            return
        if not inherits_from(clothing, ContribClothing):
            self.caller.msg(f"{clothing.name} isn't something you can wear.")
            return

        if clothing.db.worn:
            if not self.rhs:
                # If no wearstyle was provided and the clothing is already being worn, do nothing
                self.caller.msg(f"You're already wearing your {clothing.name}.")
                return
            elif len(self.rhs) > WEARSTYLE_MAXLENGTH:
                self.caller.msg(
                    f"Please keep your wear style message to less than {WEARSTYLE_MAXLENGTH} characters."
                )
                return
            else:
                # Adjust the wearstyle
                clothing.db.worn = self.rhs
                self.caller.location.msg_contents(
                    f"$You() $conj(wear) {clothing.name} {self.rhs}.", from_obj=self.caller
                )
                return

        already_worn = get_worn_clothes(self.caller)

        # Enforce overall clothing limit.
        if CLOTHING_OVERALL_LIMIT and len(already_worn) >= CLOTHING_OVERALL_LIMIT:
            self.caller.msg("You can't wear any more clothes.")
            return

        # Apply individual clothing type limits.
        if clothing_type := clothing.db.clothing_type:
            if clothing_type in CLOTHING_TYPE_LIMIT:
                type_count = single_type_count(already_worn, clothing_type)
                if type_count >= CLOTHING_TYPE_LIMIT[clothing_type]:
                    self.caller.msg(
                        f"You can't wear any more clothes of the type '{clothing_type}'."
                    )
                    return

        wearstyle = self.rhs or True
        clothing.wear(self.caller, wearstyle)


class CmdRemove(MuxCommand):
    """
    Takes off an item of clothing.

    Usage:
       remove <obj>

    Removes an item of clothing you are wearing. You can't remove
    clothes that are covered up by something else - you must take
    off the covering item first.
    """

    key = "remove"
    help_category = "clothing"

    def func(self):
        clothing = self.caller.search(self.args, candidates=self.caller.contents)
        if not clothing:
            self.caller.msg("You don't have anything like that.")
            return
        if not clothing.db.worn:
            self.caller.msg("You're not wearing that!")
            return
        if covered := clothing.db.covered_by:
            self.caller.msg(f"You have to take off {covered} first.")
            return
        clothing.remove(self.caller)


class CmdCover(MuxCommand):
    """
    Covers a worn item of clothing with another you're holding or wearing.

    Usage:
        cover <worn obj> with <obj>

    When you cover a clothing item, it is hidden and no longer appears in
    your description until it's uncovered or the item covering it is removed.
    You can't remove an item of clothing if it's covered.
    """

    key = "cover"
    help_category = "clothing"
    rhs_split = (" with ", "=")

    def func(self):
        if not len(self.args) or not self.rhs:
            self.caller.msg("Usage: cover <worn clothing> with <clothing object>")
            return

        to_cover = self.caller.search(self.lhs, candidates=get_worn_clothes(self.caller))
        cover_with = self.caller.search(self.rhs, candidates=self.caller.contents)
        if not to_cover or not cover_with:
            return
        if to_cover == cover_with:
            self.caller.msg("You can't cover an item with itself!")
            return

        if not inherits_from(cover_with, ContribClothing):
            self.caller.msg(f"{cover_with.name} isn't something you can wear.")
            return

        if cover_with.db.clothing_type in CLOTHING_TYPE_CANT_COVER_WITH:
            self.caller.msg(f"You can't cover anything with {cover_with.name}.")
            return

        if covered_by := cover_with.db.covered_by:
            self.caller.msg(f"{cover_with.name} is already covered by {covered_by.name}.")
            return
        if covered_by := to_cover.db.covered_by:
            self.caller.msg(f"{to_cover.name} is already covered by {covered_by.name}.")
            return

        # Put on the item to cover with if it's not on already
        if not cover_with.db.worn:
            cover_with.wear(self.caller, True)
        to_cover.db.covered_by = cover_with

        self.caller.location.msg_contents(
            f"$You() $conj(cover) {to_cover.name} with {cover_with.name}.", from_obj=self.caller
        )


class CmdUncover(MuxCommand):
    """
    Reveals a worn item of clothing that's currently covered up.

    Usage:
        uncover <obj>

    When you uncover an item of clothing, you allow it to appear in your
    description without having to take off the garment that's currently
    covering it. You can't uncover an item of clothing if the item covering
    it is also covered by something else.
    """

    key = "uncover"
    help_category = "clothing"

    def func(self):
        """
        This performs the actual command.
        """

        if not self.args:
            self.caller.msg("Usage: uncover <worn clothing object>")
            return

        clothing = self.caller.search(self.args, candidates=get_worn_clothes(self.caller))
        if not clothing:
            return
        if covered_by := clothing.db.covered_by:
            if covered_by.db.covered_by:
                self.caller.msg(f"{clothing.name} is under too many layers to uncover.")
                return
            clothing.db.covered_by = None
            self.caller.location.msg_contents(
                f"$You() $conj(uncover) {clothing.name}.", from_obj=self.caller
            )

        else:
            self.caller.msg(f"{clothing.name} isn't covered by anything.")
            return


class CmdInventory(MuxCommand):
    """
    view inventory

    Usage:
      inventory
      inv

    Shows your inventory.
    """

    # Alternate version of the inventory command which separates
    # worn and carried items.

    key = "inventory"
    aliases = ["inv", "i"]
    locks = "cmd:all()"
    arg_regex = r"$"

    def func(self):
        """check inventory"""
        if not self.caller.contents:
            self.caller.msg("You are not carrying or wearing anything.")
            return

        message_list = []

        items = self.caller.contents

        carry_table = evtable.EvTable(border="header")
        wear_table = evtable.EvTable(border="header")

        carried = [obj for obj in items if not obj.db.worn]
        worn = [obj for obj in items if obj.db.worn]

        message_list.append("|wYou are carrying:|n")
        for item in carried:
            carry_table.add_row(
                item.get_display_name(self.caller), item.get_display_desc(self.caller)
            )
        if carry_table.nrows == 0:
            carry_table.add_row("Nothing.", "")
        message_list.append(str(carry_table))

        message_list.append("|wYou are wearing:|n")
        for item in worn:
            item_name = item.get_display_name(self.caller)
            if item.db.covered_by:
                item_name += " (hidden)"
            wear_table.add_row(item_name, item.get_display_desc(self.caller))
        if wear_table.nrows == 0:
            wear_table.add_row("Nothing.", "")
        message_list.append(str(wear_table))

        self.caller.msg("\n".join(message_list))


class ClothedCharacterCmdSet(default_cmds.CharacterCmdSet):
    """
    Command set for clothing, including new versions of 'give' and 'drop'
    that take worn and covered clothing into account, as well as a new
    version of 'inventory' that differentiates between carried and worn
    items.
    """

    key = "DefaultCharacter"

    def at_cmdset_creation(self):
        """
        Populates the cmdset
        """
        super().at_cmdset_creation()
        #
        # any commands you add below will overload the default ones.
        #
        self.add(CmdWear())
        self.add(CmdRemove())
        self.add(CmdCover())
        self.add(CmdUncover())
        self.add(CmdInventory())



"""
Custom manager for Objects.
"""
import re

from django.conf import settings
from django.db.models import Q
from django.db.models.fields import exceptions

from evennia.server import signals
from evennia.typeclasses.managers import TypeclassManager, TypedObjectManager
from evennia.utils.utils import (
    class_from_module,
    dbid_to_obj,
    is_iter,
    make_iter,
    string_partial_matching,
)

__all__ = ("ObjectManager", "ObjectDBManager")
_GA = object.__getattribute__

# delayed import
_ATTR = None

_MULTIMATCH_REGEX = re.compile(settings.SEARCH_MULTIMATCH_REGEX, re.I + re.U)

# Try to use a custom way to parse id-tagged multimatches.


class ObjectDBManager(TypedObjectManager):
    """
    This ObjectManager implements methods for searching
    and manipulating Objects directly from the database.

    Evennia-specific search methods (will return Typeclasses or
    lists of Typeclasses, whereas Django-general methods will return
    Querysets or database objects).

    dbref (converter)
    dbref_search
    get_dbref_range
    object_totals
    typeclass_search
    get_object_with_account
    get_objs_with_key_and_typeclass
    get_objs_with_attr
    get_objs_with_attr_match
    get_objs_with_db_property
    get_objs_with_db_property_match
    get_objs_with_key_or_alias
    get_contents
    search_object (interface to many of the above methods,
                   equivalent to evennia.search_object)
    copy_object

    """

    #
    # ObjectManager Get methods
    #

    # account related

    def get_object_with_account(self, ostring, exact=True, candidates=None):
        """
        Search for an object based on its account's name or dbref.

        Args:
            ostring (str or int): Search criterion or dbref. Searching
                for an account is sometimes initiated by appending an `*` to
                the beginning of the search criterion (e.g. in
                local_and_global_search). This is stripped here.
            exact (bool, optional): Require an exact account match.
            candidates (list, optional): Only search among this list of possible
                object candidates.

        Return:
            match (query): Matching query.

        """
        ostring = str(ostring).lstrip("*")
        # simplest case - search by dbref
        dbref = self.dbref(ostring)
        if dbref:
            try:
                return self.get(db_account__id=dbref)
            except self.model.DoesNotExist:
                pass

        # not a dbref. Search by name.
        cand_restriction = (
            candidates is not None
            and Q(pk__in=[_GA(obj, "id") for obj in make_iter(candidates) if obj])
            or Q()
        )
        if exact:
            return self.filter(cand_restriction & Q(db_account__username__iexact=ostring)).order_by(
                "id"
            )
        else:  # fuzzy matching
            obj_cands = self.select_related().filter(
                cand_restriction & Q(db_account__username__istartswith=ostring)
            )
            acct_cands = [obj.account for obj in obj_cands]

            if obj_cands:
                index_matches = string_partial_matching(
                    [acct.key for acct in acct_cands], ostring, ret_index=True
                )
                acct_cands = [acct_cands[i].id for i in index_matches]
                return obj_cands.filter(db_account__id__in=acct_cands).order_by("id")

    def get_objs_with_key_and_typeclass(self, oname, otypeclass_path, candidates=None):
        """
        Returns objects based on simultaneous key and typeclass match.

        Args:
            oname (str): Object key to search for
            otypeclass_path (str): Full Python path to tyepclass to search for
            candidates (list, optional): Only match among the given list of candidates.

        Returns:
            matches (query): The matching objects.
        """
        cand_restriction = (
            candidates is not None
            and Q(pk__in=[_GA(obj, "id") for obj in make_iter(candidates) if obj])
            or Q()
        )
        return self.filter(
            cand_restriction & Q(db_key__iexact=oname, db_typeclass_path__exact=otypeclass_path)
        ).order_by("id")

    # attr/property related

    def get_objs_with_attr(self, attribute_name, candidates=None):
        """
        Get objects based on having a certain Attribute defined.

        Args:
            attribute_name (str): Attribute name to search for.
            candidates (list, optional): Only match among the given list of object
                candidates.

        Returns:
            matches (query):  All objects having the given attribute_name defined at all.

        """
        cand_restriction = (
            candidates is not None and Q(id__in=[obj.id for obj in candidates]) or Q()
        )
        return self.filter(cand_restriction & Q(db_attributes__db_key=attribute_name)).order_by(
            "id"
        )

    def get_objs_with_attr_value(
        self, attribute_name, attribute_value, candidates=None, typeclasses=None
    ):
        """
        Get all objects having the given attrname set to the given value.

        Args:
            attribute_name (str): Attribute key to search for.
            attribute_value (any):  Attribute value to search for. This can also be database
                objects.
            candidates (list, optional): Candidate objects to limit search to.
            typeclasses (list, optional): Python pats to restrict matches with.

        Returns:
            Queryset: Iterable with 0, 1 or more matches fullfilling both the `attribute_name` and
                `attribute_value` criterions.

        Notes:
            This uses the Attribute's PickledField to transparently search the database by matching
            the internal representation. This is reasonably effective but since Attribute values
            cannot be indexed, searching by Attribute key is to be preferred whenever possible.

        """
        cand_restriction = (
            candidates is not None
            and Q(pk__in=[_GA(obj, "id") for obj in make_iter(candidates) if obj])
            or Q()
        )
        type_restriction = typeclasses and Q(db_typeclass_path__in=make_iter(typeclasses)) or Q()

        results = self.filter(
            cand_restriction
            & type_restriction
            & Q(db_attributes__db_key=attribute_name)
            & Q(db_attributes__db_value=attribute_value)
        ).order_by("id")
        return results

    def get_objs_with_db_property(self, property_name, candidates=None):
        """
        Get all objects having a given db field property.

        Args:
            property_name (str): The name of the field to match for.
            candidates (list, optional): Only search among th egiven candidates.

        Returns:
            matches (list): The found matches.

        """
        property_name = "db_%s" % property_name.lstrip("db_")
        cand_restriction = (
            candidates is not None
            and Q(pk__in=[_GA(obj, "id") for obj in make_iter(candidates) if obj])
            or Q()
        )
        querykwargs = {property_name: None}
        try:
            return list(self.filter(cand_restriction).exclude(Q(**querykwargs)).order_by("id"))
        except exceptions.FieldError:
            return []

    def get_objs_with_db_property_value(
        self, property_name, property_value, candidates=None, typeclasses=None
    ):
        """
        Get objects with a specific field name and value.

        Args:
            property_name (str): Field name to search for.
            property_value (any): Value required for field with `property_name` to have.
            candidates (list, optional): List of objects to limit search to.
            typeclasses (list, optional): List of typeclass-path strings to restrict matches with

        Returns:
            Queryset: Iterable with 0, 1 or more matches.

        """
        if isinstance(property_name, str):
            if not property_name.startswith("db_"):
                property_name = "db_%s" % property_name
        querykwargs = {property_name: property_value}
        cand_restriction = (
            candidates is not None
            and Q(pk__in=[_GA(obj, "id") for obj in make_iter(candidates) if obj])
            or Q()
        )
        type_restriction = typeclasses and Q(db_typeclass_path__in=make_iter(typeclasses)) or Q()
        try:
            return self.filter(cand_restriction & type_restriction & Q(**querykwargs)).order_by(
                "id"
            )
        except exceptions.FieldError:
            return self.none()
        except ValueError:
            from evennia.utils import logger

            logger.log_err(
                "The property '%s' does not support search criteria of the type %s."
                % (property_name, type(property_value))
            )
            return self.none()

    def get_contents(self, location, excludeobj=None):
        """
        Get all objects that has a location set to this one.

        Args:
            location (Object): Where to get contents from.
            excludeobj (Object or list, optional): One or more objects
                to exclude from the match.

        Returns:
            Queryset: Iterable with 0, 1 or more matches.

        """
        exclude_restriction = (
            Q(pk__in=[_GA(obj, "id") for obj in make_iter(excludeobj)]) if excludeobj else Q()
        )
        return self.filter(db_location=location).exclude(exclude_restriction).order_by("id")

    def get_objs_with_key_or_alias(self, ostring, exact=True, candidates=None, typeclasses=None):
        """
        Args:
            ostring (str): A search criterion.
            exact (bool, optional): Require exact match of ostring
                (still case-insensitive). If `False`, will do fuzzy matching
                using `evennia.utils.utils.string_partial_matching` algorithm.
            candidates (list): Only match among these candidates.
            typeclasses (list): Only match objects with typeclasses having thess path strings.

        Returns:
            Queryset: An iterable with 0, 1 or more matches.

        """
        if not isinstance(ostring, str):
            if hasattr(ostring, "key"):
                ostring = ostring.key
            else:
                return self.none()
        if is_iter(candidates) and not len(candidates):
            # if candidates is an empty iterable there can be no matches
            # Exit early.
            return self.none()

        # build query objects
        candidates_id = [_GA(obj, "id") for obj in make_iter(candidates) if obj]
        cand_restriction = candidates is not None and Q(pk__in=candidates_id) or Q()
        type_restriction = typeclasses and Q(db_typeclass_path__in=make_iter(typeclasses)) or Q()
        if exact:
            # exact match - do direct search
            return (
                (
                    self.filter(
                        cand_restriction
                        & type_restriction
                        & (
                            Q(db_key__iexact=ostring)
                            | Q(db_tags__db_key__iexact=ostring)
                            & Q(db_tags__db_tagtype__iexact="alias")
                        )
                    )
                )
                .distinct()
                .order_by("id")
            )
        elif candidates:
            # fuzzy with candidates
            search_candidates = (
                self.filter(cand_restriction & type_restriction).distinct().order_by("id")
            )
        else:
            # fuzzy without supplied candidates - we select our own candidates
            search_candidates = (
                self.filter(
                    type_restriction
                    & (Q(db_key__icontains=ostring) | Q(db_tags__db_key__icontains=ostring))
                )
                .distinct()
                .order_by("id")
            )
        # fuzzy matching
        key_strings = search_candidates.values_list("db_key", flat=True).order_by("id")

        match_ids = []
        index_matches = string_partial_matching(key_strings, ostring, ret_index=True)
        if index_matches:
            # a match by key
            match_ids = [
                obj.id for ind, obj in enumerate(search_candidates) if ind in index_matches
            ]
        else:
            # match by alias rather than by key
            search_candidates = search_candidates.filter(
                db_tags__db_tagtype__iexact="alias", db_tags__db_key__icontains=ostring
            ).distinct()
            alias_strings = []
            alias_candidates = []
            # TODO create the alias_strings and alias_candidates lists more efficiently?
            for candidate in search_candidates:
                for alias in candidate.aliases.all():
                    alias_strings.append(alias)
                    alias_candidates.append(candidate)
            index_matches = string_partial_matching(alias_strings, ostring, ret_index=True)
            if index_matches:
                # it's possible to have multiple matches to the same Object, we must weed those out
                match_ids = [alias_candidates[ind].id for ind in index_matches]
        # TODO - not ideal to have to do a second lookup here, but we want to return a queryset
        # rather than a list ... maybe the above queries can be improved.
        return self.filter(id__in=match_ids)

    # main search methods and helper functions

    def search_object(
        self,
        searchdata,
        attribute_name=None,
        typeclass=None,
        candidates=None,
        exact=True,
        use_dbref=True,
        tags=None,
    ):
        """
        Search as an object globally or in a list of candidates and
        return results. The result is always an Object. Always returns
        a list.

        Args:
            searchdata (str or Object): The entity to match for. This is
                usually a key string but may also be an object itself.
                By default (if no `attribute_name` is set), this will
                search `object.key` and `object.aliases` in order.
                Can also be on the form #dbref, which will (if
                `exact=True`) be matched against primary key.
            attribute_name (str): Use this named Attribute to
                match searchdata against, instead of the defaults. If
                this is the name of a database field (with or without
                the `db_` prefix), that will be matched too.
            typeclass (str or TypeClass): restrict matches to objects
                having this typeclass. This will help speed up global
                searches.
            candidates (list): If supplied, search will
                only be performed among the candidates in this list. A
                common list of candidates is the contents of the
                current location searched.
            exact (bool): Match names/aliases exactly or partially.
                Partial matching matches the beginning of words in the
                names/aliases, using a matching routine to separate
                multiple matches in names with multiple components (so
                "bi sw" will match "Big sword"). Since this is more
                expensive than exact matching, it is recommended to be
                used together with the `candidates` keyword to limit the
                number of possibilities. This value has no meaning if
                searching for attributes/properties.
            use_dbref (bool): If False, bypass direct lookup of a string
                on the form #dbref and treat it like any string.
            tags (list): A list of tuples `(tagkey, tagcategory)` where the
                matched object must have _all_ tags in order to be considered
                a match.

        Returns:
            matches (list): Matching objects

        """

        def _searcher(searchdata, candidates, typeclass, exact=False):
            """
            Helper method for searching objects. `typeclass` is only used
            for global searching (no candidates)
            """
            if attribute_name:
                # attribute/property search (always exact).
                matches = self.get_objs_with_db_property_value(
                    attribute_name, searchdata, candidates=candidates, typeclasses=typeclass
                )
                if matches:
                    return matches
                return self.get_objs_with_attr_value(
                    attribute_name, searchdata, candidates=candidates, typeclasses=typeclass
                )
            else:
                # normal key/alias search
                return self.get_objs_with_key_or_alias(
                    searchdata, exact=exact, candidates=candidates, typeclasses=typeclass
                )

        def _search_by_tag(query, taglist):
            if not query:
                query = self.all()

            for tagkey, tagcategory in taglist:
                query = query.filter(db_tags__db_key=tagkey, db_tags__db_category=tagcategory)

            return query

        if not searchdata and searchdata != 0:
            if tags:
                return _search_by_tag(make_iter(tags))

            return self.none()

        if typeclass:
            # typeclass may also be a list
            typeclasses = make_iter(typeclass)
            for i, typeclass in enumerate(make_iter(typeclasses)):
                if callable(typeclass):
                    typeclasses[i] = "%s.%s" % (typeclass.__module__, typeclass.__name__)
                else:
                    typeclasses[i] = "%s" % typeclass
            typeclass = typeclasses

        if candidates is not None:
            if not candidates:
                # candidates is the empty list. This should mean no matches can ever be acquired.
                return []
            # Convenience check to make sure candidates are really dbobjs
            candidates = [cand for cand in make_iter(candidates) if cand]
            if typeclass:
                candidates = [
                    cand for cand in candidates if _GA(cand, "db_typeclass_path") in typeclass
                ]

        dbref = not attribute_name and exact and use_dbref and self.dbref(searchdata)
        if dbref:
            # Easiest case - dbref matching (always exact)
            dbref_match = self.dbref_search(dbref)
            if dbref_match:
                dmatch = dbref_match[0]
                if not candidates or dmatch in candidates:
                    return dbref_match
                else:
                    return self.none()

        # Search through all possibilities.
        match_number = None
        # always run first check exact - we don't want partial matches
        # if on the form of 1-keyword etc.
        matches = _searcher(searchdata, candidates, typeclass, exact=True)

        if not matches:
            # no matches found - check if we are dealing with N-keyword
            # query - if so, strip it.
            match = _MULTIMATCH_REGEX.match(str(searchdata))
            match_number = None
            stripped_searchdata = searchdata
            if match:
                # strips the number
                match_number, stripped_searchdata = match.group("number"), match.group("name")
                match_number = int(match_number) - 1
            if match_number is not None:
                # run search against the stripped data
                matches = _searcher(stripped_searchdata, candidates, typeclass, exact=True)
                if not matches:
                    # final chance to get a looser match against the number-strippped query
                    matches = _searcher(stripped_searchdata, candidates, typeclass, exact=False)
            elif not exact:
                matches = _searcher(searchdata, candidates, typeclass, exact=False)

        if tags:
            matches = _search_by_tag(matches, make_iter(tags))

        # deal with result
        if len(matches) == 1 and match_number is not None and match_number != 0:
            # this indicates trying to get a single match with a match-number
            # targeting some higher-number match (like 2-box when there is only
            # one box in the room). This leads to a no-match.
            matches = self.none()
        elif len(matches) > 1 and match_number is not None:
            # multiple matches, but a number was given to separate them
            if 0 <= match_number < len(matches):
                # limit to one match (we still want a queryset back)
                # TODO: Can we do this some other way and avoid a second lookup?
                matches = self.filter(id=matches[match_number].id)
            else:
                # a number was given outside of range. This means a no-match.
                matches = self.none()

        # return a list (possibly empty)
        return matches

    # alias for backwards compatibility
    object_search = search_object
    search = search_object

    #
    # ObjectManager Copy method

    def copy_object(
        self,
        original_object,
        new_key=None,
        new_location=None,
        new_home=None,
        new_permissions=None,
        new_locks=None,
        new_aliases=None,
        new_destination=None,
    ):
        """
        Create and return a new object as a copy of the original object. All
        will be identical to the original except for the arguments given
        specifically to this method. Object contents will not be copied.

        Args:
            original_object (Object): The object to make a copy from.
            new_key (str, optional): Name of the copy, if different
                from the original.
            new_location (Object, optional): Alternate location.
            new_home (Object, optional): Change the home location
            new_aliases (list, optional): Give alternate object
                aliases as a list of strings.
            new_destination (Object, optional): Used only by exits.

        Returns:
            copy (Object or None): The copy of `original_object`,
                optionally modified as per the ingoing keyword
                arguments.  `None` if an error was encountered.

        """

        # get all the object's stats
        typeclass_path = original_object.typeclass_path
        if not new_key:
            new_key = original_object.key
        if not new_location:
            new_location = original_object.location
        if not new_home:
            new_home = original_object.home
        if not new_aliases:
            new_aliases = original_object.aliases.all()
        if not new_locks:
            new_locks = original_object.db_lock_storage
        if not new_permissions:
            new_permissions = original_object.permissions.all()
        if not new_destination:
            new_destination = original_object.destination

        # create new object
        from evennia.scripts.models import ScriptDB
        from evennia.utils import create

        new_object = create.create_object(
            typeclass_path,
            key=new_key,
            location=new_location,
            home=new_home,
            permissions=new_permissions,
            locks=new_locks,
            aliases=new_aliases,
            destination=new_destination,
        )
        if not new_object:
            return None

        # copy over all attributes from old to new.
        attrs = (
            (a.key, a.value, a.category, a.lock_storage) for a in original_object.attributes.all()
        )
        new_object.attributes.batch_add(*attrs)

        # copy over all cmdsets, if any
        for icmdset, cmdset in enumerate(original_object.cmdset.all()):
            if icmdset == 0:
                new_object.cmdset.add_default(cmdset)
            else:
                new_object.cmdset.add(cmdset)

        # copy over all scripts, if any
        for script in original_object.scripts.all():
            ScriptDB.objects.copy_script(script, new_obj=new_object)

        # copy over all tags, if any
        tags = (
            (t.db_key, t.db_category, t.db_data) for t in original_object.tags.all(return_objs=True)
        )
        new_object.tags.batch_add(*tags)

        return new_object

    def clear_all_sessids(self):
        """
        Clear the db_sessid field of all objects having also the
        db_account field set.

        """
        self.filter(db_sessid__isnull=False).update(db_sessid=None)

    def create_object(
        self,
        typeclass=None,
        key=None,
        location=None,
        home=None,
        permissions=None,
        locks=None,
        aliases=None,
        tags=None,
        destination=None,
        report_to=None,
        nohome=False,
        attributes=None,
        nattributes=None,
    ):
        """

        Create a new in-game object.

        Keyword Args:
            typeclass (class or str): Class or python path to a typeclass.
            key (str): Name of the new object. If not set, a name of
                `#dbref` will be set.
            location (Object or str): Obj or #dbref to use as the location of the new object.
            home (Object or str): Obj or #dbref to use as the object's home location.
            permissions (list): A list of permission strings or tuples (permstring, category).
            locks (str): one or more lockstrings, separated by semicolons.
            aliases (list): A list of alternative keys or tuples (aliasstring, category).
            tags (list): List of tag keys or tuples (tagkey, category) or (tagkey, category, data).
            destination (Object or str): Obj or #dbref to use as an Exit's target.
            report_to (Object): The object to return error messages to.
            nohome (bool): This allows the creation of objects without a
                default home location; only used when creating the default
                location itself or during unittests.
            attributes (list): Tuples on the form (key, value) or (key, value, category),
                (key, value, lockstring) or (key, value, lockstring, default_access).
                to set as Attributes on the new object.
            nattributes (list): Non-persistent tuples on the form (key, value). Note that
                adding this rarely makes sense since this data will not survive a reload.

        Returns:
            object (Object): A newly created object of the given typeclass.

        Raises:
            ObjectDB.DoesNotExist: If trying to create an Object with
                `location` or `home` that can't be found.

        """
        typeclass = typeclass if typeclass else settings.BASE_OBJECT_TYPECLASS

        # convenience converters to avoid common usage mistake
        permissions = make_iter(permissions) if permissions is not None else None
        locks = make_iter(locks) if locks is not None else None
        aliases = make_iter(aliases) if aliases is not None else None
        tags = make_iter(tags) if tags is not None else None
        attributes = make_iter(attributes) if attributes is not None else None

        if isinstance(typeclass, str):
            # a path is given. Load the actual typeclass
            typeclass = class_from_module(typeclass, settings.TYPECLASS_PATHS)

        # Setup input for the create command. We use ObjectDB as baseclass here
        # to give us maximum freedom (the typeclasses will load
        # correctly when each object is recovered).

        location = dbid_to_obj(location, self.model)
        destination = dbid_to_obj(destination, self.model)

        if home:
            home_obj_or_dbref = home
        elif nohome:
            home_obj_or_dbref = None
        else:
            home_obj_or_dbref = settings.DEFAULT_HOME

        try:
            home = dbid_to_obj(home_obj_or_dbref, self.model)
        except self.model.DoesNotExist:
            if settings.TEST_ENVIRONMENT:
                # this happens for databases where the #1 location is flushed during tests
                home = None
            else:
                raise self.model.DoesNotExist(
                    f"settings.DEFAULT_HOME (= '{settings.DEFAULT_HOME}') does not exist, "
                    "or the setting is malformed."
                )

        # create new instance
        new_object = typeclass(
            db_key=key,
            db_location=location,
            db_destination=destination,
            db_home=home,
            db_typeclass_path=typeclass.path,
        )
        # store the call signature for the signal
        new_object._createdict = dict(
            key=key,
            location=location,
            destination=destination,
            home=home,
            typeclass=typeclass.path,
            permissions=permissions,
            locks=locks,
            aliases=aliases,
            tags=tags,
            report_to=report_to,
            nohome=nohome,
            attributes=attributes,
            nattributes=nattributes,
        )
        # this will trigger the save signal which in turn calls the
        # at_first_save hook on the typeclass, where the _createdict can be
        # used.
        new_object.save()

        signals.SIGNAL_OBJECT_POST_CREATE.send(sender=new_object)

        return new_object


class ObjectManager(ObjectDBManager, TypeclassManager):
    pass



"""
Exits

Exits are connectors between Rooms. An exit always has a destination property
set and has a single command defined on itself with the same name as its key,
for allowing Characters to traverse the exit to its destination.

"""
from evennia.objects.objects import DefaultExit

from .objects import ObjectParent


class Exit(ObjectParent, DefaultExit):
    """
    Exits are connectors between rooms. Exits are normal Objects except
    they defines the `destination` property. It also does work in the
    following methods:

     basetype_setup() - sets default exit locks (to change, use `at_object_creation` instead).
     at_cmdset_get(**kwargs) - this is called when the cmdset is accessed and should
                              rebuild the Exit cmdset along with a command matching the name
                              of the Exit object. Conventionally, a kwarg `force_init`
                              should force a rebuild of the cmdset, this is triggered
                              by the `@alias` command when aliases are changed.
     at_failed_traverse() - gives a default error message ("You cannot
                            go there") if exit traversal fails and an
                            attribute `err_traverse` is not defined.

    Relevant hooks to overload (compared to other types of Objects):
        at_traverse(traveller, target_loc) - called to do the actual traversal and calling of the other hooks.
                                            If overloading this, consider using super() to use the default
                                            movement implementation (and hook-calling).
        at_post_traverse(traveller, source_loc) - called by at_traverse just after traversing.
        at_failed_traverse(traveller) - called by at_traverse if traversal failed for some reason. Will
                                        not be called if the attribute `err_traverse` is
                                        defined, in which case that will simply be echoed.
    """

    pass



"""
Roleplaying emotes and language - Griatch, 2015

"""

from .rplanguage import LanguageExistsError  # noqa
from .rplanguage import LanguageHandler  # noqa
from .rplanguage import (
    LanguageError,
    add_language,
    available_languages,
    obfuscate_language,
    obfuscate_whisper,
)
from .rpsystem import CmdSay  # noqa
from .rpsystem import ContribRPCharacter  # noqa
from .rpsystem import ContribRPObject  # noqa
from .rpsystem import ContribRPRoom  # noqa
from .rpsystem import RPSystemCmdSet  # noqa
from .rpsystem import (
    CmdEmote,
    CmdMask,
    CmdPose,
    CmdRecog,
    CmdSdesc,
    EmoteError,
    LanguageError,
    RecogError,
    RecogHandler,
    RPCommand,
    SdescError,
    SdescHandler,
    parse_language,
    parse_sdescs_and_recogs,
    send_emote,
)



"""
Protfuncs are FuncParser-callables that can be embedded in a prototype to provide custom logic
without having access to Python. The protfunc is parsed at the time of spawning, using the creating
object's session as input. If the protfunc returns a non-string, this is what will be added to the
prototype.

In the prototype dict, the protfunc is specified as a string inside the prototype, e.g.:

```python
{ ...

"key": "$funcname(args, kwargs)"

...  }
```

Available protfuncs are either all callables in one of the modules of `settings.PROT_FUNC_MODULES`
or all callables added to a dict FUNCPARSER_CALLABLES in such a module. By default, base inlinefuncs
for text manipulation and searching are included, as well as the special `$protkey` function.
See the Prototypes and Spawner documentation for more info.

```python
def funcname (*args, **kwargs):
    return "replacement text"
```

At spawn-time the spawner passes the following extra kwargs into each callable (in addition to
what is added in the call itself):

- `session` (Session): The Session of the entity spawning using this prototype.
- `prototype` (dict): The dict this protfunc is a part of.
- `current_key` (str): The active key this value belongs to in the prototype.

Any traceback raised by this function will be handled at the time of spawning and abort the spawn
before any object is created/updated. It must otherwise return the value to store for the specified
prototype key (this value must be possible to serialize in an Attribute).

"""

from evennia.utils import funcparser


def protfunc_callable_protkey(*args, **kwargs):
    """
    Usage: $protkey(keyname)

    Returns the value of another key in this prototoype. Will raise an error if
    the key is not found in this prototype.

    """
    if not args:
        return ""

    prototype = kwargs.get("prototype", {})
    fieldname = args[0]
    prot_value = None
    if fieldname in prototype:
        prot_value = prototype[fieldname]
    else:
        # check if it's an attribute
        for attrtuple in prototype.get("attrs", []):
            if attrtuple[0] == fieldname:
                prot_value = attrtuple[1]
                break
        else:
            raise AttributeError(
                f"{fieldname} not found in prototype\n{prototype}\n"
                "(neither as prototype-field or as an Attribute"
            )
    if callable(prot_value):
        raise RuntimeError(
            f"Error in prototype\n{prototype}\n$protkey can only reference static "
            f"values/attributes (found {prot_value})"
        )
    try:
        return funcparser.funcparser_callable_eval(prot_value, **kwargs)
    except funcparser.ParsingError:
        return prot_value


# this is picked up by FuncParser
FUNCPARSER_CALLABLES = {
    "protkey": protfunc_callable_protkey,
    **funcparser.FUNCPARSER_CALLABLES,
    **funcparser.SEARCHING_CALLABLES,
}



"""
Object

The Object is the "naked" base class for things in the game world.

Note that the default Character, Room and Exit does not inherit from
this Object, but from their respective default implementations in the
evennia library. If you want to use this class as a parent to change
the other types, you can do so by adding this as a multiple
inheritance.

"""
from evennia.objects.objects import DefaultObject


class ObjectParent:
    """
    This is a mixin that can be used to override *all* entities inheriting at
    some distance from DefaultObject (Objects, Exits, Characters and Rooms).

    Just add any method that exists on `DefaultObject` to this class. If one
    of the derived classes has itself defined that same hook already, that will
    take precedence.

    """


class Object(ObjectParent, DefaultObject):
    """
    This is the root typeclass object, implementing an in-game Evennia
    game object, such as having a location, being able to be
    manipulated or looked at, etc. If you create a new typeclass, it
    must always inherit from this object (or any of the other objects
    in this file, since they all actually inherit from BaseObject, as
    seen in src.object.objects).

    The BaseObject class implements several hooks tying into the game
    engine. By re-implementing these hooks you can control the
    system. You should never need to re-implement special Python
    methods, such as __init__ and especially never __getattribute__ and
    __setattr__ since these are used heavily by the typeclass system
    of Evennia and messing with them might well break things for you.


    * Base properties defined/available on all Objects

     key (string) - name of object
     name (string)- same as key
     dbref (int, read-only) - unique #id-number. Also "id" can be used.
     date_created (string) - time stamp of object creation

     account (Account) - controlling account (if any, only set together with
                       sessid below)
     sessid (int, read-only) - session id (if any, only set together with
                       account above). Use `sessions` handler to get the
                       Sessions directly.
     location (Object) - current location. Is None if this is a room
     home (Object) - safety start-location
     has_account (bool, read-only)- will only return *connected* accounts
     contents (list of Objects, read-only) - returns all objects inside this
                       object (including exits)
     exits (list of Objects, read-only) - returns all exits from this
                       object, if any
     destination (Object) - only set if this object is an exit.
     is_superuser (bool, read-only) - True/False if this user is a superuser

    * Handlers available

     aliases - alias-handler: use aliases.add/remove/get() to use.
     permissions - permission-handler: use permissions.add/remove() to
                   add/remove new perms.
     locks - lock-handler: use locks.add() to add new lock strings
     scripts - script-handler. Add new scripts to object with scripts.add()
     cmdset - cmdset-handler. Use cmdset.add() to add new cmdsets to object
     nicks - nick-handler. New nicks with nicks.add().
     sessions - sessions-handler. Get Sessions connected to this
                object with sessions.get()
     attributes - attribute-handler. Use attributes.add/remove/get.
     db - attribute-handler: Shortcut for attribute-handler. Store/retrieve
            database attributes using self.db.myattr=val, val=self.db.myattr
     ndb - non-persistent attribute handler: same as db but does not create
            a database entry when storing data

    * Helper methods (see src.objects.objects.py for full headers)

     search(ostring, global_search=False, attribute_name=None,
             use_nicks=False, location=None, ignore_errors=False, account=False)
     execute_cmd(raw_string)
     msg(text=None, **kwargs)
     msg_contents(message, exclude=None, from_obj=None, **kwargs)
     move_to(destination, quiet=False, emit_to_obj=None, use_destination=True)
     copy(new_key=None)
     delete()
     is_typeclass(typeclass, exact=False)
     swap_typeclass(new_typeclass, clean_attributes=False, no_default=True)
     access(accessing_obj, access_type='read', default=False)
     check_permstring(permstring)

    * Hooks (these are class methods, so args should start with self):

     basetype_setup()     - only called once, used for behind-the-scenes
                            setup. Normally not modified.
     basetype_posthook_setup() - customization in basetype, after the object
                            has been created; Normally not modified.

     at_object_creation() - only called once, when object is first created.
                            Object customizations go here.
     at_object_delete() - called just before deleting an object. If returning
                            False, deletion is aborted. Note that all objects
                            inside a deleted object are automatically moved
                            to their <home>, they don't need to be removed here.

     at_init()            - called whenever typeclass is cached from memory,
                            at least once every server restart/reload
     at_cmdset_get(**kwargs) - this is called just before the command handler
                            requests a cmdset from this object. The kwargs are
                            not normally used unless the cmdset is created
                            dynamically (see e.g. Exits).
     at_pre_puppet(account)- (account-controlled objects only) called just
                            before puppeting
     at_post_puppet()     - (account-controlled objects only) called just
                            after completing connection account<->object
     at_pre_unpuppet()    - (account-controlled objects only) called just
                            before un-puppeting
     at_post_unpuppet(account) - (account-controlled objects only) called just
                            after disconnecting account<->object link
     at_server_reload()   - called before server is reloaded
     at_server_shutdown() - called just before server is fully shut down

     at_access(result, accessing_obj, access_type) - called with the result
                            of a lock access check on this object. Return value
                            does not affect check result.

     at_pre_move(destination)             - called just before moving object
                        to the destination. If returns False, move is cancelled.
     announce_move_from(destination)         - called in old location, just
                        before move, if obj.move_to() has quiet=False
     announce_move_to(source_location)       - called in new location, just
                        after move, if obj.move_to() has quiet=False
     at_post_move(source_location)          - always called after a move has
                        been successfully performed.
     at_object_leave(obj, target_location)   - called when an object leaves
                        this object in any fashion
     at_object_receive(obj, source_location) - called when this object receives
                        another object

     at_traverse(traversing_object, source_loc) - (exit-objects only)
                              handles all moving across the exit, including
                              calling the other exit hooks. Use super() to retain
                              the default functionality.
     at_post_traverse(traversing_object, source_location) - (exit-objects only)
                              called just after a traversal has happened.
     at_failed_traverse(traversing_object)      - (exit-objects only) called if
                       traversal fails and property err_traverse is not defined.

     at_msg_receive(self, msg, from_obj=None, **kwargs) - called when a message
                             (via self.msg()) is sent to this obj.
                             If returns false, aborts send.
     at_msg_send(self, msg, to_obj=None, **kwargs) - called when this objects
                             sends a message to someone via self.msg().

     return_appearance(looker) - describes this object. Used by "look"
                                 command by default
     at_desc(looker=None)      - called by 'look' whenever the
                                 appearance is requested.
     at_get(getter)            - called after object has been picked up.
                                 Does not stop pickup.
     at_drop(dropper)          - called when this object has been dropped.
     at_say(speaker, message)  - by default, called if an object inside this
                                 object speaks

    """

    pass



"""
The Django Rest Framework provides a way of generating urls for different
views that implement standard CRUD operations in a quick way, using 'routers'
and 'viewsets'. A viewset implements standard CRUD actions and any custom actions
that you want, and then a router will automatically generate URLs based on the
actions that it detects for a viewset. For example, below we create a DefaultRouter.
We then register ObjectDBViewSet, a viewset for CRUD operations for ObjectDB
instances, to the 'objects' base endpoint. That will generate a number of URLs
like the following:

list objects:    action: GET, url: /objects/, view name: object-list
create object:   action: POST, url: /objects/, view name: object-list
retrieve object: action: GET, url: /objects/<:pk>, view name: object-detail
update object:   action: POST, url: /objects/<:pk>, view name: object-detail
delete object:   action: DELETE, url: /objects/<:pk>, view name: object-detail
set attribute:   action: POST, url: /objects/<:pk>/set-attribute, view name: object-set-attribute

"""

from django.urls import path
from django.views.generic import TemplateView
from rest_framework.schemas import get_schema_view

from evennia.web.api import views
from evennia.web.api.root import APIRootRouter

app_name = "api"

router = APIRootRouter()
router.trailing_slash = "/?"
router.register(r"accounts", views.AccountDBViewSet, basename="account")
router.register(r"objects", views.ObjectDBViewSet, basename="object")
router.register(r"characters", views.CharacterViewSet, basename="character")
router.register(r"exits", views.ExitViewSet, basename="exit")
router.register(r"rooms", views.RoomViewSet, basename="room")
router.register(r"scripts", views.ScriptDBViewSet, basename="script")
router.register(r"helpentries", views.HelpViewSet, basename="helpentry")

urlpatterns = router.urls

urlpatterns += [
    # openapi schema
    path(
        "openapi",
        get_schema_view(title="Evennia API", description="Evennia OpenAPI Schema", version="1.0"),
        name="openapi",
    ),
    # redoc auto-doc (based on openapi schema)
    path(
        "redoc/",
        TemplateView.as_view(
            template_name="rest_framework/redoc.html", extra_context={"schema_url": "api:openapi"}
        ),
        name="redoc",
    ),
]



"""
Main test-suite runner of Evennia. The runner collates tests from
all over the code base and runs them.

Runs as part of the Evennia's test suite with 'evennia test evennia"

"""
from django.test.runner import DiscoverRunner


class EvenniaTestSuiteRunner(DiscoverRunner):
    """
    Pointed to by the TEST_RUNNER setting.
    This test runner only runs tests on the apps specified in evennia/
     avoid running the large number of tests defined by Django

    """

    def setup_test_environment(self, **kwargs):
        import evennia

        evennia._init()

        from django.conf import settings

        # set testing flag while suite runs
        settings.TEST_ENVIRONMENT = True
        super().setup_test_environment(**kwargs)

    def teardown_test_environment(self, **kwargs):
        # remove testing flag after suite has run
        from django.conf import settings

        settings.TEST_ENVIRONMENT = False

        super().teardown_test_environment(**kwargs)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.9 on 2016-09-21 17:31


from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("comms", "0008_auto_20160905_0902")]

    operations = [
        migrations.AlterField(
            model_name="msg",
            name="db_hide_from_channels",
            field=models.ManyToManyField(
                blank=True, null=True, related_name="hide_from_channels_set", to="comms.ChannelDB"
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_hide_from_objects",
            field=models.ManyToManyField(
                blank=True, null=True, related_name="hide_from_objects_set", to="objects.ObjectDB"
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_hide_from_accounts",
            field=models.ManyToManyField(
                blank=True,
                null=True,
                related_name="hide_from_accounts_set",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_receivers_channels",
            field=models.ManyToManyField(
                blank=True,
                help_text="channel recievers",
                null=True,
                related_name="channel_set",
                to="comms.ChannelDB",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_receivers_objects",
            field=models.ManyToManyField(
                blank=True,
                help_text="object receivers",
                null=True,
                related_name="receiver_object_set",
                to="objects.ObjectDB",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_receivers_accounts",
            field=models.ManyToManyField(
                blank=True,
                help_text="account receivers",
                null=True,
                related_name="receiver_account_set",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_sender_external",
            field=models.CharField(
                blank=True,
                db_index=True,
                help_text="identifier for external sender, for example a sender over an IRC connection (i.e. someone who doesn't have an exixtence in-game).",
                max_length=255,
                null=True,
                verbose_name="external sender",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_sender_objects",
            field=models.ManyToManyField(
                blank=True,
                db_index=True,
                null=True,
                related_name="sender_object_set",
                to="objects.ObjectDB",
                verbose_name="sender(object)",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_sender_accounts",
            field=models.ManyToManyField(
                blank=True,
                db_index=True,
                null=True,
                related_name="sender_account_set",
                to=settings.AUTH_USER_MODEL,
                verbose_name="sender(account)",
            ),
        ),
        migrations.AlterField(
            model_name="msg",
            name="db_tags",
            field=models.ManyToManyField(
                blank=True,
                help_text="tags on this message. Tags are simple string markers to identify, group and alias messages.",
                null=True,
                to="typeclasses.Tag",
            ),
        ),
    ]



"""
Test gendersub contrib.

"""


from mock import patch

from evennia.commands.default.tests import BaseEvenniaCommandTest
from evennia.utils.create import create_object

from . import gendersub


class TestGenderSub(BaseEvenniaCommandTest):
    def test_setgender(self):
        self.call(gendersub.SetGender(), "male", "Your gender was set to male.")
        self.call(gendersub.SetGender(), "ambiguous", "Your gender was set to ambiguous.")
        self.call(gendersub.SetGender(), "Foo", "Usage: @gender")

    def test_gendercharacter(self):
        char = create_object(gendersub.GenderCharacter, key="Gendered", location=self.room1)
        txt = "Test |p gender"
        self.assertEqual(
            gendersub._RE_GENDER_PRONOUN.sub(char._get_pronoun, txt), "Test their gender"
        )
        with patch(
            "evennia.contrib.game_systems.gendersub.gendersub.DefaultCharacter.msg"
        ) as mock_msg:
            char.db.gender = "female"
            char.msg(txt)
            mock_msg.assert_called_with("Test her gender", from_obj=None, session=None)

    def test_gendering_others(self):
        """ensure characters see the gender of the sender, not themselves"""
        fem = create_object(
            gendersub.GenderCharacter,
            key="Gendered",
            location=self.room2,
            attributes=[("gender", "female")],
        )
        masc = create_object(
            gendersub.GenderCharacter,
            key="Gendered",
            location=self.room2,
            attributes=[("gender", "male")],
        )
        txt = "Test |p gender"

        with patch(
            "evennia.contrib.game_systems.gendersub.gendersub.DefaultCharacter.msg"
        ) as mock_msg:
            fem.msg(txt, from_obj=masc)
            self.assertIn("Test his gender", mock_msg.call_args.args)
            masc.msg(txt, from_obj=fem)
            self.assertIn("Test her gender", mock_msg.call_args.args)

    def test_ungendered_source(self):
        char = create_object(gendersub.GenderCharacter, key="Gendered", location=self.room1)
        txt = "Test |p gender"
        with patch(
            "evennia.contrib.game_systems.gendersub.gendersub.DefaultCharacter.msg"
        ) as mock_msg:
            char.msg(txt, from_obj=self.char1)
            self.assertIn("Test their gender", mock_msg.call_args.args)



import re

from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

from evennia.accounts.models import AccountDB


class EvenniaUsernameAvailabilityValidator:
    """
    Checks to make sure a given username is not taken or otherwise reserved.
    """

    def __call__(self, username):
        """
        Validates a username to make sure it is not in use or reserved.

        Args:
            username (str): Username to validate

        Returns:
            None (None): None if password successfully validated,
                raises ValidationError otherwise.

        """
        # Check guest list
        if settings.GUEST_LIST and username.lower() in (
            guest.lower() for guest in settings.GUEST_LIST
        ):
            raise ValidationError(
                _("Sorry, that username is reserved."), code="evennia_username_reserved"
            )

        # Check database
        exists = AccountDB.objects.filter(username__iexact=username).exists()
        if exists:
            raise ValidationError(
                _("Sorry, that username is already taken."), code="evennia_username_taken"
            )
